### CVM WEB SERVICE - BRAZILLIAN SEC

import backoff
import pandas as pd
import multiprocessing as mp
from requests import request
from requests.exceptions import ReadTimeout
from getpass import getuser
from time import sleep
from lxml import html
from typing import Tuple, Optional
from random import shuffle
from stpstone.handling_data.html import HtmlHndler, SeleniumWD
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.str import StrHandler
from stpstone.loggs.create_logs import CreateLog
from stpstone.handling_data.folders import DirFilesManagement
from stpstone.handling_data.lists import HandlingLists
from stpstone.multithreading.mp_helper import mp_worker, mp_run_parallel
from stpstone.documents_numbers.br import DocumentsNumbersBR


class CVMWeb_WS_Funds:

    def __init__(
        self,
        str_cookie:Optional[str]=None,
        str_id:Optional[str]=None,
        str_passw:Optional[str]=None,
        webdriver_path:Optional[str]=None,
        port:Optional[str]=None,
        time_wait_page_load:Optional[int]=10,
        time_wait_el:Optional[int]=10,
        bl_open_minimized:Optional[bool]=True,
        bl_parallel:bool=False,
        bl_insert_or_ignore:bool=True,
        cls_db:type=None,
        int_sleep:object=1,
        int_ncpus:int=mp.cpu_count() - 2 if mp.cpu_count() > 2 else 1,
        key_fund_code:str='fund_code',
        key_fund_daily_infos_url:str='url',
        key_fund_ein:str='fund_ein',
        key_fund_name:str='fund_name',
        key_ref_date:str='ref_date',
        key_total_portfolio:str='total_portfolio',
        key_aum:str='aum',
        key_quote:str='quote',
        key_fund_raising:str='fund_raising',
        key_redemptions:str='redemptions',
        key_provisioned_redemptions:str='provisioned_redemptions',
        key_liquid_assets:str='liquid_assets',
        key_num_shareholders:str='num_shareholders',
        key_fund_ein_unm='fund_ein_unmasked',
        fstr_greatest_shareholders:str='greatest_shareholder_{}',
        key_greatest_shareholders_like:str='greatest_shareholders_*',
        str_host_ex_fund:str=r'https://cvmweb.cvm.gov.br/SWB/Sistemas/SCW/CReservd/',
        str_host_post_fund:str=r'https://cvmweb.cvm.gov.br/SWB/Sistemas/ADM/CReservada/InfDiario/'
    ):
        '''
        DOCSTRING: CVM WEB SERVICE - BRAZILLIAN SEC INPUTS
        INPUTS: COOKIE CAN BE CATCH TRACKING NETWORK WHEN LOGGING WITH THE GOV.BR ACCOUNT IN 
            https://cvmweb.cvm.gov.br/swb/default.asp?sg_sistema=scw (SEARCH FOR ouvidor=0)
        OUTPUTS: - 
        '''
        self.str_id = str_id
        self.str_passw = str_passw
        self.webdriver_path = webdriver_path
        self.port = port
        self.time_wait_page_load = time_wait_page_load
        self.time_wait_el = time_wait_el
        self.bl_open_minimized = bl_open_minimized
        self.str_cookie = str_cookie if str_cookie is not None else self.cookie_govbr(
            str_id,
            str_passw,
            webdriver_path,
            port,
            time_wait_page_load,
            time_wait_el,
            bl_open_minimized
        )
        self.bl_parallel = bl_parallel
        self.bl_insert_or_ignore = bl_insert_or_ignore
        self.cls_db = cls_db
        self.int_sleep = int_sleep
        self.int_ncpus = int_ncpus
        self.dict_cookie = {'Cookie': self.str_cookie}
        self.key_fund_code = key_fund_code
        self.key_fund_daily_infos_url = key_fund_daily_infos_url
        self.key_fund_ein = key_fund_ein
        self.key_fund_name = key_fund_name
        self.key_ref_date = key_ref_date
        self.key_total_portfolio = key_total_portfolio
        self.key_aum = key_aum
        self.key_quote = key_quote
        self.key_fund_raising = key_fund_raising
        self.key_redemptions = key_redemptions
        self.key_provisioned_redemptions = key_provisioned_redemptions
        self.key_liquid_assets = key_liquid_assets
        self.key_num_shareholders = key_num_shareholders
        self.key_fund_ein_unm = key_fund_ein_unm
        self.fstr_greatest_shareholders = fstr_greatest_shareholders
        self.key_greatest_shareholders_like = key_greatest_shareholders_like
        self.str_host_ex_fund = str_host_ex_fund
        self.str_host_post_fund = str_host_post_fund

    def cookie_govbr(self, 
        str_id:str,
        str_passw:str,
        webdriver_path:str,
        port:str,
        time_wait_page_load:int=30,
        time_wait_el:int=30,
        bl_open_minimized:bool=False,
        url:str='https://cvmweb.cvm.gov.br/swb/default.asp?sg_sistema=scw', 
        str_frame_name:str='Main',
        xpath_click_login:str='//*[@id="linkGovBr"]/img',
        xpath_fill_tin_br:str='//*[@id="accountId"]',
        xpath_fill_passw:str='//*[@id="password"]',
    ) -> str:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        cls_selenium = SeleniumWD()
        # create driver element, wait until page is loaded and the login button is available
        driver = cls_selenium.selenium_web_driver(url, webdriver_path, port, time_wait_page_load, 
            bl_open_minimized)
        cls_selenium.selenium_element_is_clickable(driver, xpath_click_login, 
            frame_name=str_frame_name)
        # click login button
        cls_selenium.selenium_find_element(driver, xpath_click_login).click()
        # fill TIN br - tax payer id number
        cls_selenium.selenium_wait_until_element_loaded(driver, xpath_fill_tin_br, time_wait_el)
        cls_selenium.selenium_find_element(driver, xpath_fill_tin_br).send_keys(str_id)
        # fill password
        cls_selenium.selenium_wait_until_element_loaded(driver, xpath_fill_passw, time_wait_el)
        cls_selenium.selenium_find_element(driver, xpath_fill_passw).send_keys(str_passw)
        # get cookie
        list_events = cls_selenium.get_network_traffic(driver)
        print(list_events)
        raise Exception('BREAK')

    def generic_req(self, str_method:str, url:str, str_header_ref:str, dict_data:dict={}, 
        bl_allow_redirects:bool=True, tup_timeout:Tuple[float, float]=(12.0, 21.0)) -> html.HtmlElement:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        dict_headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'Accept-Language': 'en-US,en;q=0.9,pt;q=0.8,es;q=0.7',
            'Cache-Control': 'max-age=0',
            'Connection': 'keep-alive',
            'Cookie': f'{self.str_cookie}',
            'Referer': 'https://cvmweb.cvm.gov.br/SWB/Sistemas/SCW/CReservd/{}'.format(str_header_ref),
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'same-origin',
            'Sec-Fetch-User': '?1',
            'Upgrade-Insecure-Requests': '1',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36',
            'sec-ch-ua': '"Google Chrome";v="131", "Chromium";v="131", "Not_A Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"'
        }
        resp_req = request(
            method=str_method, 
            url=url, 
            headers=dict_headers, 
            data=dict_data, 
            cookies=self.dict_cookie,
            allow_redirects=bl_allow_redirects,
            timeout=tup_timeout
        )
        resp_req.raise_for_status()
        html_content = HtmlHndler().html_lxml_parser(page=resp_req.text)
        return html_content

    @property
    @backoff.on_exception(
        backoff.constant,
        ReadTimeout,
        interval=10,
        max_tries=20,
    )
    def available_funds(
        self, 
        str_table_nane:str='RAW_CVMWEB_COD_FUNDS',
        str_header_ref:str='CReservd.asp',
        str_app:str='SelecPartic.aspx?CD_TP_INFORM=15',
        str_xpath_funds:str='//select[@id="PK_PARTIC"]/option',
        str_xpath_value:str='./@value',
        str_method:str='GET',
        bl_allow_redirects:bool=True
    ) -> pd.DataFrame:
        '''
        DOCSTRING: AVAILABLE FUND CODES AND EIN
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_ser = list()
        # request html
        html_content = self.generic_req(str_method, self.str_host_ex_fund + str_app, 
                                        str_header_ref, bl_allow_redirects=bl_allow_redirects)
        # print(f'HTML CONTENT: \n{html_content}')
        # print('OPTION FUND: {}'.format(HtmlHndler().html_lxml_xpath(html_content, '//select[@id="PK_PARTIC"]/option[3]/text()')))
        # looping within available funds and filling serialized list
        for el_option_fund in HtmlHndler().html_lxml_xpath(html_content, str_xpath_funds):
            #   get text and split into fund's ein and name
            try:
                str_option_fund_raw = HtmlHndler().html_lxml_xpath(el_option_fund, './text()')
                if len(str_option_fund_raw) == 0: 
                    continue
                else:
                    str_option_fund_raw = str_option_fund_raw[0]
                str_option_fund_trt = StrHandler().remove_diacritics(str_option_fund_raw)
                str_option_fund_trt = StrHandler().replace_all(
                    str_option_fund_trt,
                    {
                        '- (CLASSE DE COTAS) -': '- (CLASSE DE COTAS)',
                        '- CREDITO PRIVADO': 'CREDITO PRIVADO',
                        '- RENDA FIXA -': 'RENDA FIXA',
                        'FUNDO DE INVESTIMENTO - RENDA FIXA': 'FUNDO DE INVESTIMENTO RENDA FIXA',
                        '- FUNDO DE INVESTIMENTO EM COTAS DE FUNDOS DE INVESTIMENTO': \
                            'FUNDO DE INVESTIMENTO EM COTAS DE FUNDOS DE INVESTIMENTO',
                        '- FGTS ': 'FGTS ',
                        '- CRED PRIV': 'CRED PRIV',
                        '- INVESTIMENTO NO EXTERIOR': 'INVESTIMENTO NO EXTERIOR',
                        '- FUNDO DE INVESTIMENTO EM ACOES': 'FUNDO DE INVESTIMENTO EM ACOES',
                        ' - IE': 'IE',
                        '- RESPONSABILIDADE LIMITADA': 'RESPONSABILIDADE LIMITADA',
                    }
                )
                str_fund_ein, str_fund_name =  str_option_fund_trt.split(' - ')
            except ValueError:
                raise Exception(
                    '** ERROR: FUND EIN AND NAME NOT FOUND \n\t' 
                    + '- COMMON ERROR IS AN UNEXPECTED "-" CHARACTER, PROVIDED THE CODE'
                    + ' EXPECTS ONLY ONE SEPARATOR FOR FUND EIN AND NAME \n\t'
                    + '- RAW DROP DOWN OPTION FUND: {} \n\t'.format(str_option_fund_raw)
                    + '- REPLACED OPTION FUND: {} \n\t'.format(str_option_fund_trt)
                    + '- URL CVMWEB: {}'.format(self.str_host_ex_fund + str_app)
                )
            #   backup in db, if is user's will
            list_ser.append({
                self.key_fund_code: HtmlHndler().html_lxml_xpath(
                    el_option_fund, str_xpath_value)[0],
                self.key_fund_ein: str_fund_ein,
                self.key_fund_name: str_fund_name,
            })
        # creating pandas dataframe
        df_funds = pd.DataFrame(list_ser)
        # changing columns types
        df_funds = df_funds.astype({
            self.key_fund_code: str,
            self.key_fund_ein: str,
            self.key_fund_name: str
        })
        # unmasking ein
        df_funds[self.key_fund_ein_unm] = DocumentsNumbersBR(
            df_funds[self.key_fund_ein].tolist()).unmask_docs
        # loading in db, if is user's will
        if self.cls_db is not None: 
            self.cls_db._insert(
                df_funds.to_dict(orient='records'), 
                str_table_nane, 
                bl_insert_or_ignore=self.bl_insert_or_ignore
            )
        # returning dataframe
        return df_funds

    def fund_daily_report_gen(
        self,
        str_fund_code:str,
        dict_data:dict={},
        str_header_ref:str='SelecPartic.aspx?CD_TP_INFORM=15',
        str_app:str='ConsInfDiario.aspx?PK_PARTIC={}&PK_SUBCLASSE=-1',
        str_method:str='GET',
        bl_allow_redirects:bool=False,
    ) -> Tuple[html.HtmlElement, str]:
        '''
        DOCSTRING: GENERAL FUND DAILY REPORT WITH THE MOST RECENT DATE
        INPUTS: FUND CODE
        OUTPUTS: HTML
        '''
        # print(f'URL: {self.str_host_post_fund + str_app.format(str_fund_code)}')
        return \
            self.generic_req(
                str_method, 
                self.str_host_post_fund + str_app.format(str_fund_code), 
                str_header_ref, 
                dict_data,
                bl_allow_redirects=bl_allow_redirects,
            ), \
            self.str_host_post_fund + str_app.format(str_fund_code)

    def available_dates_report_fund(
        self,
        html_content:html.HtmlElement,
        str_fund_code:str,
        url_fund_daily_infos:str,
        str_table_nane:str='RAW_CVMWEB_DATAS_FUNDOS',
        str_dt_fmt:str='DD/MM/YYYY',
        xpath_avl_dates:str='//*[@id="ddCOMPTC"]/option/text()'
    ):
        '''
        DOCSTRING: AVAILABLE DATES FOR FUND DAY REPORT
        INPUTS: FUND CODE
        OUTPUTS: LIST
        '''
        list_dts = [
            el_ 
            for el_ in HtmlHndler().html_lxml_xpath(html_content, xpath_avl_dates) 
            if el_ != ''
        ]
        # print(f'**\nFUND_CODE: {str_fund_code} / LIST_DATAS_CVM: {list_dts}')
        list_avl_dts_fund = [
            {
                self.key_fund_code: str_fund_code,
                self.key_ref_date: DatesBR().str_date_to_datetime(d, str_dt_fmt),
                self.key_fund_daily_infos_url: url_fund_daily_infos,
            } for d in list_dts
        ]
        if \
            (self.cls_db is not None) \
            and (len(list_avl_dts_fund) > 0):
            self.cls_db._insert(
                list_avl_dts_fund, 
                str_table_nane, 
                bl_insert_or_ignore=self.bl_insert_or_ignore
            )
        return list_dts

    def fund_daily_reports_raw(
        self, 
        html_content_gen:str,
        str_fund_code:str,
        list_str_dts:list,
        el_eventtarget='__EVENTTARGET',
        el_eventargument='__EVENTARGUMENT',
        el_lastfocus='__LASTFOCUS',
        el_viewstate='__VIEWSTATE',
        el_eventvalidation='__EVENTVALIDATION',
        el_viewstategen='__VIEWSTATEGENERATOR',
        str_xpath_event_el_like:str='//*[@id="{}"]/@value',
        str_event_target:str='ddCOMPTC',
        str_header_ref:str='SelecPartic.aspx?CD_TP_INFORM=15',
        str_app:str='ConsInfDiario.aspx?PK_PARTIC={}&PK_SUBCLASSE=-1',
        str_method_fund_report_dt:str='POST',
        bl_allow_redirects:bool=False
    ):
        '''
        DOCSTRING: SIMULATION OF JAVASCRIPT CODE EXECUTION WHEN THE REFERENCE DATE IS CHANGED IN THE 
            DROP DOWN MENU, IN ORDER TO RETRIEVE THE FUND DAILY REPORTS FOR DATES OF INTEREST
        INPUTS: 
        OUTPUTS:
        '''
        # setting variables
        dict_html_contents_dts = dict()
        # looping within list of dates
        # print(f'LIST_STR_DTS:{list_str_dts}')
        for str_dt in list_str_dts:
            #   build data dictionary
            dict_data = {
                el_eventtarget: str_event_target,
                el_eventargument: '',
                el_lastfocus: '',
                el_viewstate: HtmlHndler().html_lxml_xpath(html_content_gen, 
                    str_xpath_event_el_like.format(el_viewstate)),
                el_eventvalidation: HtmlHndler().html_lxml_xpath(html_content_gen, 
                    str_xpath_event_el_like.format(el_eventvalidation)),
                el_viewstategen: HtmlHndler().html_lxml_xpath(html_content_gen, str_xpath_event_el_like.format(
                    el_viewstategen)),
                str_event_target: str_dt
            }
            # print(f'PARAMS FORM INFOS DIARIAS CVM: {dict_data}')
            # print(f'DICT_PARAMS_DAILY_INFOS: {dict_data}')
            #   request html for a specific date of fund report and serialize it
            # html_content = self.generic_req(
            #     str_method_fund_report_dt, 
            #     self.str_host_post_fund + str_app.format(str_fund_code), 
            #     str_header_ref, 
            #     dict_data,
            #     bl_allow_redirects=bl_allow_redirects,
            # )
            resp_req = request(
                method=str_method_fund_report_dt, 
                url=self.str_host_post_fund + str_app.format(str_fund_code),
                allow_redirects=False, 
                data=dict_data, 
                cookies=self.dict_cookie
            )
            resp_req.raise_for_status()
            html_content = HtmlHndler().html_lxml_parser(page=resp_req.text)
            # print(f'\nLIST_SER2_DAILY_INFOS: {list_ser_2}')
            dict_html_contents_dts[str_dt] = html_content
        # return html contents
        return dict_html_contents_dts

    def fund_daily_report_trt(
        self, 
        url_daily_report_fund:str,
        html_content_gen:str,
        str_fund_code:str,
        list_str_dts:list,
        dict_xpaths:dict={
            'total_portfolio': '//*[@id="lblTotalCarteira"]/text()',
            'aum': '//*[@id="lblValorPL"]/text()',
            'quote': '///*[@id="lbVlrCota"]/text()',
            'fund_raising': '//*[@id="lblVlrCaptacoes"]/text()',
            'redemptions': '//*[@id="lblVlrResgates"]/text()',
            'provisioned_redemptions': '//*[@id="lblVlrTotalSaidas"]/text()',
            'liquid_assets': '//*[@id="lblValorTotalAtivos"]/text()',
            'num_shareholders': '//*[@id="lblNumCotistas"]/text()'
        },
        str_xpath_greatest_shareholders:str='//*[starts-with(@id, "lblCNPJPartic")]/text()',
        str_xpath_dropdown_box:str='//*[@id="ddCOMPTC"]/option[1]/@value',
        str_fmt_date:str='DD/MM/YYYY'
    ) -> list:
        # setting variables
        list_ser = list()
        # in case of no available dates, return an empty list
        # print('DROPDOWN_EL: {}'.format(
        #     len(HtmlHndler().html_lxml_xpath(html_content_gen, 
        #         str_xpath_dropdown_box)) == 0
        # ))
        # print(f'URL: {url_daily_report_fund}')
        # raise Exception('BREAK')
        if \
            (len(list_str_dts) == 0) \
            or (len(HtmlHndler().html_lxml_xpath(html_content_gen, 
                str_xpath_dropdown_box)) == 0): 
            return []
        # daily reports for a given set of fund and dates
        dict_html_contents_dts = self.fund_daily_reports_raw(
            html_content_gen,
            str_fund_code,
            list_str_dts
        )
        # print(f'HTML INFOS DIARIAS - DTS CONSULTAR: {dict_html_contents_dts}')
        # looping within available funds and filling serialized list
        for str_dt, html_content_dt in dict_html_contents_dts.items():
            #   greatest shareholders and their holdings in the current fund
            dict_g_shareholders = {
                self.fstr_greatest_shareholders.format(i): el_ein[0] if el_ein is not None else ''
                for i, el_ein in enumerate(
                    HtmlHndler().html_lxml_xpath(html_content_dt, str_xpath_greatest_shareholders)
                )
            }
            #   daily infos
            try:
                dict_daily_infos = {
                    self.__dict__[f'key_{k}']: str(
                        HtmlHndler().html_lxml_xpath(html_content_dt, v)[0]).replace(',', '.') 
                    for k, v in dict_xpaths.items()
                }
            except IndexError:
                raise Exception(f'ERROR HTML XML XPATH - URL: {url_daily_report_fund}')
            dict_daily_infos[self.key_fund_code] = str_fund_code
            dict_daily_infos[self.key_fund_daily_infos_url] = url_daily_report_fund
            dict_daily_infos[self.key_ref_date] = DatesBR().str_date_to_datetime(str_dt, str_fmt_date)
            #   appending to exporting list
            dict_ = HandlingDicts().merge_n_dicts(dict_daily_infos, dict_g_shareholders)
            # print(f'DICT_DAILY_INFOS: {dict_}')
            list_ser.append(dict_)
        # retuning dictionary
        return list_ser

    def block_fund_fetch(
        self,
        str_fund_code:str,
        dict_dts_funds:dict,
        str_code_version:str='dev',
        str_table_nane:str='RAW_CVMWEB_INFOS_DIARIAS'
    ) -> list:
        '''
        DOCSTRING: BLOCK FUND DAILY REPORTS FOR DATES OF INTEREST TO FETCH - PARALLELIZED
        INPUTS: FUND CODE, DATES OF INTEREST, DIRECTORY TO BACKUP, NAME OF BACKUP FILE, CODE VERSION
        OUTPUTS: LIST OF DICTIONARIES
        '''
        # generic request for the given fund code
        html_content_gen, url_fund_daily_infos = self.fund_daily_report_gen(str_fund_code)
        # dates of daily report available
        list_str_dts_reports = self.available_dates_report_fund(html_content_gen, 
            str_fund_code, url_fund_daily_infos)
        # filtering dates of interest with daily reports available
        list_ftd_dts = [
            d 
            for d in dict_dts_funds[str_fund_code] 
            if d in list_str_dts_reports
        ]
        # print(f'\n** FUND_CODE: {str_fund_code} / LIST_FLTD_DTS: {list_ftd_dts}')
        # retrieving available daily reports for the given funds, bounded by dates of interest
        list_ser = self.fund_daily_report_trt(
            url_fund_daily_infos, html_content_gen, str_fund_code, list_ftd_dts)
        # print(f'LIST_SER_CVMWEB_DAILY_INFOS: \n{list_ser}')
        # backup in db, if is user's will
        if \
            (self.cls_db is not None) \
            and (len(list_ser) > 0):
            self.cls_db._insert(
                list_ser, 
                str_table_nane, 
                bl_insert_or_ignore=self.bl_insert_or_ignore
            )
        # wait for the next iteration, if is user's will
        if self.int_sleep is not None:
            sleep(self.int_sleep)
        # returning list of dictionaries
        return list_ser

    def funds_daily_reports_trt(
        self,
        dict_dts_funds:dict,
        str_code_version:str='dev',
        str_dt_fmt_1:str='YYYY-MM-DD',
        str_strftime_format:str='%d/%m/%Y',
    ) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_funds = list()
        # check date format
        for str_fund_code, list_dts in dict_dts_funds.items():
            for i, str_dt in enumerate(list_dts):
                if DatesBR().check_date_datetime_format(str_dt) == True:
                    list_dts[i] = str_dt.strftime(str_dt_fmt_1)
                elif \
                    (StrHandler().match_string_like(str_dt, '*-*') == True) \
                    and (isinstance(str_dt, str) == True):
                    list_dts[i] = DatesBR().str_date_to_datetime(
                        str_dt, format=str_dt_fmt_1).strftime(str_strftime_format)
            dict_dts_funds[str_fund_code] = list_dts
        #   parallelized fetch, if is user's will
        if self.bl_parallel == True:
            #   preparing task arguments
            list_task_args = [
                (
                    (str_fund_code, dict_dts_funds, str_code_version), 
                    {}
                )
                for str_fund_code in list(dict_dts_funds.keys())
            ]
            #   executing tasks in parallel
            list_funds = mp_run_parallel(
                self.block_fund_fetch(
                    str_fund_code,
                    dict_dts_funds,
                    str_code_version
                ), 
                list_task_args, 
                int_ncpus=self.int_ncpus
            )
            #   flattening list
            list_funds = HandlingLists().flatten_list(list_funds)
        else:
            # randomize fund codes
            list_fnds_cds = list(dict_dts_funds.keys())
            shuffle(list_fnds_cds)
            # loop within funds
            for str_fund_code in list_fnds_cds:
                list_funds.extend(
                    self.block_fund_fetch(
                        str_fund_code,
                        dict_dts_funds,
                        str_code_version
                    )
                )
        # appending to pandas dataframe
        df_funds_daily_reports = pd.DataFrame(list_funds)
        # print(df_funds_daily_reports)
        # changing data types
        df_funds_daily_reports = df_funds_daily_reports.astype({
            self.key_fund_code: str,
            self.key_total_portfolio: float,
            self.key_aum: float,
            self.key_quote: float,
            self.key_fund_raising: float,
            self.key_redemptions: float,
            self.key_provisioned_redemptions: float,
            self.key_liquid_assets: float,
            self.key_num_shareholders: int,
            self.key_fund_daily_infos_url: str
        })
        cols_greatest_shareholders = [
            c 
            for c in df_funds_daily_reports.columns 
            if StrHandler().match_string_like(c, self.key_greatest_shareholders_like) == True
        ]
        df_funds_daily_reports[cols_greatest_shareholders] = df_funds_daily_reports[
            cols_greatest_shareholders
        ].astype(str)
        # remove duplicated data
        df_funds_daily_reports.drop_duplicates(inplace=True)
        # retuning dataframe
        return df_funds_daily_reports


### ENVIO DE MENSAGENS AUTOMÁTICAS NO TEAMS ###

import pymsteams


class WebhookTeams:

    def __init__(self, url_webhook:str) -> None:
        self.url_webhook = url_webhook

    def send_message(self, str_msg:str, str_title:str='ROUTINE_CONCLUSION', 
                           bl_print_message:bool=False) -> None:
        '''
        DOCSTRING: SEND PLAIN MESSAGE WITH TEXT AND TITLE
        INPUTS: WEBHOOK CONNECTION, MESSAGE, TITLE AND BODY
        OUTPUTS: -
        '''
        teams_message = pymsteams.connectorcard(self.url_webhook)
        teams_message.title(str_title)
        teams_message.text(str_msg)
        if bl_print_message == True:
            teams_message.printme()
        teams_message.send()



### GLOBAL CONSTANTS FOR HARDCODES USAGE

import os
from stpstone.opening_config.setup import reading_yaml

# slots of memory to each yaml
YAML_ANBIMA = reading_yaml(r'{}\anbima.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_B3 = reading_yaml(r'{}\b3.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_BR_TRS = reading_yaml(r'{}\br_treasury.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_CD = reading_yaml(r'{}\comdinheiro.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_GEN = reading_yaml(r'{}\generic.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_WGBD = reading_yaml(r'{}\world_gov_bonds.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_GLB_RT = reading_yaml(r'{}\global_rates.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_USA_MACRO = reading_yaml(r'{}\usa_macro.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_BR_MACRO = reading_yaml(r'{}\br_macro.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_MICROSOFT_APPS = reading_yaml(r'{}\microsoft_apps.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_LLMS = reading_yaml(r'{}\llms.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))
YAML_SESSION = reading_yaml(r'{}\session.yaml'.format(
    os.path.dirname(os.path.realpath(__file__))))


### STATISTICAL HYPOTHESIS TESTS ###

import numpy as np
import statsmodels.api as sm
import statsmodels.stats.diagnostic as dg
from scipy import stats
stats.chisqprob = lambda chisq, df: stats.chi2.sf(chisq, df)
from sklearn.linear_model import LinearRegression
from statsmodels.stats.stattools import durbin_watson
from stpstone.quantitative_methods.statistical_description import StatiscalDescription
from stpstone.quantitative_methods.prob_distributions import NormalDistribution
from stpstone.quantitative_methods.eda import ExploratoryDataAnalysis


class MultipleRegressionHT:

    def anova(self, array_x, array_y):
        '''
        REFERENCES: https://medium.com/swlh/interpreting-linear-regression-through-statsmodels-summary-4796d359035a
        DOCSTRING: ANOVA - ANALYSIS OF VARIANCE
        INPUTS:
        OUTPUTS:
        '''
        # fitting multiple regression model
        model_fitted = sm.OLS(array_y, array_x).fit()
        # returning stats
        return model_fitted.summary()

    def linearity_test(self, array_x, array_y, list_cols_iv, r_squared_cut=0.8):
        '''
        DOCSTRING: LINEARITY TEST - LINEAR RELATIONSHIP BETWEEN DV AND EACH IV
        INPUTS: ARRAY X, ARRAY Y, R-SQUARED CUT
        OUTPUTS: DICTIONARY
        '''
        # setting variables
        list_linearity = list()
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # looping through each iv
        for idx in range(array_x.shape[1]):
            # fit a simple linear regression model
            lr_model = np.polyfit(array_x[:, idx], array_y, 1)
            # r squared caclulation
            float_r_squared = 1.0 - np.var(array_y - np.polyval(lr_model, array_x[:, idx])) \
                / np.var(array_x[:, idx])
            list_linearity.append(float_r_squared)
        # print(list_linearity)
        return {
            'r_squared': {list_cols_iv[idx]: list_linearity[idx] for idx in 
                          range(len(list_linearity))},
            'h0': 'linear relationship between X and y', 
            'bl_reject_h0': {list_cols_iv[idx]: list_linearity[idx] < r_squared_cut for idx in 
                          range(len(list_linearity))}
        }

    def vif_iv(self, vector_x, array_y):
        '''
        DOCSTRING: VARIANCE INFLATION FACTOR - INDEPENDENCE OF INDEPENDENT VARIABLES - VIOLATION: 
            MULTICOLLINEARITY; ISSUE: INFLATED STANDARD ERRORS; DETECTION: VARIANCE INFLATION 
            FACTOR; CORRECTION: REVISE MODEL, INCREASE SAMPLE SIZE
        INPUTS
        '''
        # fit a simple linear regression model
        lr_model = np.polyfit(vector_x, array_y, 1)
        # r squared caclulation
        float_r_squared = 1.0 - np.var(array_y - np.polyval(lr_model, vector_x)) / np.var(vector_x)
        # calculate individual vif
        float_vif = 1.0 / (1.0 - float_r_squared)
        # return vif
        return float_vif, float_r_squared
    
    def calculate_vif_ivs(self, array_x, array_y, list_cols_iv, float_r_squared_mc_cut=0.8):
        '''
        DOCSTRING: VARIANCE INFLATION FACTOR - IF VIF SCORE FOR A VARIABLE X_J IS HIGH, THEN THE 
            VARIANCE EXPLAINED BY THAT VARIABLE IS CAPTURED (ALMOST COMPLETELY) BY ALL OTHER 
            VARIABLES --> TEST FOR MULTICOLLINEARITY ASSUMPTION OF LINEAR REGRESSION
        INPUTS: ARRAY INDEPENDENT AND DEPENDENT
        OUTPUTS:
        '''
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # iterating through array iv columns to calculate vif regarding one iv and the dv
        tups_vif_ivs = [self.vif_iv(array_x[:, idx], array_y) for idx in range(array_x.shape[1])]
        # return vif for ivs
        return {
            'vif_ivs': {col_: tups_vif_ivs[idx][0] for idx, col_ in enumerate(list_cols_iv)}, 
            'r_squared_mc_ivs': {col_: tups_vif_ivs[idx][1] for idx, col_ in 
                                 enumerate(list_cols_iv)}, 
            'h0': 'lack of multicollinearity',
            'bl_reject_h0': {col_: tups_vif_ivs[idx][0] > float_r_squared_mc_cut for idx, col_ in 
                                 enumerate(list_cols_iv)}
        }

    def normality_error_dist_test(self, array_x, array_y, alpha=0.05):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # fitting multiple regression model
        model_fitted = sm.OLS(array_y, array_x).fit()
        # anderson darling normality check of error's distribution
        dict_anderson_darling = NormalityHT().anderson_darling(model_fitted.resid, alpha)
        return dict_anderson_darling

    def breusch_godfrey(self, array_x, array_y, p=None):
        '''
        REFERENCES: https://en.wikipedia.org/wiki/Breusch%E2%80%93Godfrey_test
        DOCSTRING: BREUSCH GODFREY TEST ASSESS INDEPENDENCE (ABSENSE OF SERIAL CORRELATION, OR 
            AUTOCORRELATION); IF REJECTED, IT WOULD MEAN THAT INCORRECT CONCLUSIONS WOULD BE DRAWN 
            FROM OTHER TESTS OR THAT SUB-OPTIMAL ESTIMATES OF MODEL PARAMETERS WOULD BE OBTAINED
            - THE REGRESSION MODELS TO WHICH THE TEST CAN BE APPLIED INCLUDE CASES WHERE LAGGED 
                VALUESOF THE DEPENDENT VARIABLES ARE USED AS INDEPENDENT VARIABLES IN THE MODEL'S 
                REPRESENTATION FOR LATER OBSERVATIONS
            - TEST FOR AUTOCORRELATION IN THE ERRORS OF A REGRESSION MODEL
            - H0: NO SERIAL CORRELATION OF ANY ORDER UP TO P
            - HA: SERIAL CORRELATION OF ANY ORDER UP TO P
        INPUTS: ARRAY INDEPENDENT, DEPENDENT
        OUTPUTS: DICTIONARY WITH BREUSH GODFREY TEST OUTPUT (TUPPLE, BEING THE FIRST VALUE THE TEST 
            STATISTIC AND THE SECOND THE P-VALUE, WHICH OUGHT BE LESS THAN THE TEST STATISTIC TO 
            ACCEPT THE H0 HYPOTHESIS)
        '''
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # defining the order p of the test
        if p == None:
            p = array_x.shape[1] + 1
        # fitting multiple regression model
        model_fitted = sm.OLS(array_y, array_x).fit()
        # autocorrelation test for time-series residuals at order p
        tup_dg = dg.acorr_breusch_godfrey(model_fitted, nlags=p)
        # return dictionary
        return {
            'breush_godfrey_tup': tup_dg,
            'h0': 'no serial correlation of any order up to p', 
            'bl_reject_h0': tup_dg[0] < tup_dg[1]
        }

    def durbin_watson_test(self, array_x, array_y):
        '''
        REFERENCES: https://www.statology.org/durbin-watson-test-python/
        DOCSTRING: DURBIN-WATSON SERIAL CORRELATION: ONE OF THE ASSUMPTIONS OF LINEAR REGRESSION 
            IS THAT THERE IS NO CORRELATION BETWEEN THE RESIDUALS (ASSUMED TO BE INDEPENDENT), 
            ONE WAY TO DETERMINE WHETER THIS ASSUMPTION IS MET IS TO PERFORM A 
            DURBIN-WATSON TEST, WHICH IS USED TO DETECT THE PRESENCE OF AUTOCORRELATION IN 
            THE RESIDUALS OF A REGRESSION
            A. H0: NO CORRELATION AMONG THE RESIDUALS
            B. HA: THE RESIDUALS ARE AUTOCORRELATED
            C. VALUES: 0 TO 4, WHERE A TEST STATISTIC OF 2 INDICATES NO SERIAL CORRELATION, 
                THE CLOSER THE TEST STATISTIC IS TO 0, THE MORE EVIDENCE OF POSITIVE SERIAL 
                CORRELATION, THE CLOSER THE TEST STATISTIC IS TO 4, THE MORE EVIDENCE OF NEGATIVE 
                SERIAL CORRELATION
            D. RULE OF THUMB: TEST STATISTIC VALUES BETWEEN 1.5 AND 2.5 ARE CONSIDERED NORMAL, 
                HOWEVER, VALUES OUTSIDE OF THIS RANGE COULD INDICATE THAT AUTOCORRELATION IS A 
                PROBLEM
            E. DURBIN-WATSON IS A TEST OF SERIAL CORRELATION  
        INPUTS: ARRAY INDEPENDENT AND DEPENDENT
        OUTPUTS: DICTIONARY
        '''
        # fitting multiple regression model
        model_fitted = sm.OLS(array_y, array_x).fit()
        # durbin watson test value
        float_dw_test_value = durbin_watson(model_fitted.resid)
        # test
        return {
            'test_value': float_dw_test_value, 
            'h0': 'no autocorrelation among the residuals',
            'bl_reject_h0': (float_dw_test_value <= 1.5) or (float_dw_test_value >= 2.5)
        }

    def breusch_pagan_test(self, array_x, array_y, alpha=0.05):
        '''
        REFERENCES: https://stackoverflow.com/questions/30061054/ols-breusch-pagan-test-in-python
        DOCSTRING: BREUSCH-PAGAN: TEST WHETER HETEROSCEDASTICITY EXISTS IN REGRESSION ANALYSIS, DEFINED 
            AS THE UNEQUAL SCATTERING OF RESIDUALS
            A. H0: HOMOSCEDASTICITY - EQUAL VARIANCE AMONG RESIDUALS
            B. HA: HETEROSKEDASTICITY - UNEQUAL VARIANCE AMONG RESIDUALS
        INPUTS:
            X = A NUMPY.NDARRAY CONTAINING THE PREDICTOR VARIABLES. SHAPE = (NSAMPLES, NPREDICTORS).
            Y = A 1D NUMPY.NDARRAY CONTAINING THE RESPONSE VARIABLE. SHAPE = (NSAMPLES, ).
        OUTPUTS: A DICTIONARY CONTAINING THREE ELEMENTS
            1. THE BREUSCH-PAGAN TEST STATISTIC.
            2. THE P-VALUE FOR THE TEST.
            3. THE TEST RESULT.
        '''
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # checking dv and ivs
        if array_y.ndim != 1:
            raise SystemExit('Error: array_y has more than 1 dimension.')
        if array_x.shape[0] != array_y.shape[0]:
            raise SystemExit('Error: the number of samples differs between array_x and array_y.')
        else:
            n_samples = array_y.shape[0]
        # fit an OLS linear model to array_y using array_x
        lm = LinearRegression()
        lm.fit(array_x, array_y)
        # calculate the squared errors
        err = (array_y - lm.predict(array_x))**2
        # fit an auxiliary regression to the squared errors to estimate the variance 
        #   in err explained by array_x
        lm.fit(array_x, err)
        pred_err = lm.predict(array_x)
        del lm
        # calculate the coefficient of determination
        ss_tot = sum((err - np.mean(err))**2)
        ss_res = sum((err - pred_err)**2)
        r2 = 1 - (ss_res / ss_tot)
        del err, pred_err, ss_res, ss_tot
        # calculate the Lagrange multiplier:
        lagrange_multiplier = n_samples * r2
        del r2
        # calculate p-value. degrees of freedom = number of predictors.
        # this is equivalent to (p - 1) parameter restrictions in Wikipedia entry.
        pval = stats.chisqprob(lagrange_multiplier, array_x.shape[1])
        # hipothesis test for heteroskedasticity
        bl_resject_h0 = alpha > pval
        # returning results
        return {
            'lagrange_multiplier': lagrange_multiplier, 
            'p_value': pval, 
            'h0': 'homoscedasticity',
            'bl_reject_h0': bl_resject_h0
        }
    
    def cooks_distance(self, array_x, array_y):
        '''
        DOCSTRING: COOK'S DISTANCE DETECTS OUTLIERS, WHICH EXCLUSION MAY CAUSE SUBSTANTIAL 
            CHANGES IN THE ESTIMATED REGRESSION FUNCTION - METRIC FOR IDENTIFYING INFLUENTIAL 
            DATA POINTS
            A. TUPLE OF ARRAYS WITH COOK'S DISTANCE AND P-VALUE
            B. LARGE COOK'S D INFLUENCE: 
                B.1. D > 0.5 MAY BE INFLUENTIAL
                B.2. D > 1 LIKELY TO BE INFLUENTIAL
                B.3. D > 2 / SQRT(K / N) - LARGE SAMPLES
        INPUTS:
        OUTPUTS:
        '''
        # fitting multiple regression model
        model_fitted = sm.OLS(array_y, array_x).fit()
        # calculate influence and outlier measures
        influence = model_fitted.get_influence()
        # return cook's distance
        return influence.cooks_distance

    def test_joint_coeff(self, sse_r, sse_unr, q, n, k):
        '''
        DOCSTRING: F-STAT TEST FOR JOINT COEFFICIENTS
        INPUTS: SSE RESTRICTED, SSE UNRESTRICTED, Q (NUMBER OF RESTRICTIONS), 
            N (NUMBER OF OBSERVATIONS), K (NUMBER OF INDEPENDENT VARIABLES)
        OUTPUTS: FLOAT
        '''
        return (sse_r - sse_unr) / q / (sse_unr / (n - k - 1))

class NormalityHT:

    def kolmogorov_smirnov_test(self, array_x, alpha=0.05, factor_empirical_func=10):
        '''
        REFERENCES: http://www.portalaction.com.br/inferencia/62-teste-de-kolmogorov-smirnov
        DOCSTRING: TEST OF HYPOTHESIS KOLMOGOROV-SMIRNOV TO CHECK WHETER THE NORMAL DISTRIBUTIONS
            FITS A SAMPLE OF DATA
        INPUTS: LIST OF REAL NUMBERS, ALPHA (STANDARD VALUE IS 0.05) AND
            FACTOR FOR EMPIRICAL FUNCTION
        OUTPUTS: DICTIONARY WITH DN, CRITICAL VALUE AND WHETHER TO REJECT OR NOT H0 HYPOTHESIS THAT
            THE DATA IS FITTED BY A NORMAL DISTRIBUTION
        '''
        # sort list
        array_x.sort()
        # turning list to numpy array
        array_x = np.array(array_x)
        # empirical function array
        empirical_func_array = [
            array_x / factor_empirical_func for array_x in range(1, len(array_x) + 1)]
        # theoretical function array, with normal function distribution for cumulative values
        mean_array = StatiscalDescription().statistical_description(
            array_x)['mean']
        standard_deviation = StatiscalDescription().statistical_description(array_x)[
            'standard_deviation_sample']
        cdf_array = [NormalDistribution().cdf(array_x, mean_array, standard_deviation)
                     for array_x in np.array(array_x)]
        # test 1 array: |F(array_x(i)) - Fn(array_x(i-1))|
        test_1_array = [abs(empirical_func_array[i] - cdf_array[i])
                        for i in range(len(array_x))]
        # test 2 array: |F(array_x(i))-Fn(array_x(i-1))|
        test_2_array = list()
        for i in range(len(array_x)):
            if i == 0:
                test_2_array.append(cdf_array[i])
            else:
                test_2_array.append(cdf_array[i] - empirical_func_array[i - 1])
        test_2_array = np.array(test_2_array)
        # either reject or not h0 hypothesis
        return {
            'dn': max(np.array(np.max(test_1_array)), np.array(np.max(test_2_array))),
            'critical_value': stats.ksone.ppf(
                1 - alpha / 2, len(array_x)),
            'h0': 'nomally distributed data',
            'reject_h0': max(np.array(np.max(test_1_array)), np.array(np.max(
                test_2_array))) > stats.ksone.ppf(1 - alpha / 2, len(array_x))
        }

    def anderson_darling(self, array_data, alpha=0.05):
        '''
        REFERENCES: http://www.portalaction.com.br/inferencia/63-teste-de-anderson-darling,
            http://www.uel.br/projetos/experimental/pages/arquivos/Anderson_Darling.html
        DOCSTRING: TEST OF HYPOTHESIS ANDERSON-DARLING TO CHECK WHETER THE NORMAL DISTRIBUTIONS
            FITS A SAMPLE OF DATA
        INPUTS: LIST OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        array_data.sort()
        # turning list to numpy array
        array_x = np.array(array_data)
        # statistical description
        mean_array = StatiscalDescription().statistical_description(
            array_x)['mean']
        standard_deviation = StatiscalDescription().statistical_description(array_x)[
            'standard_deviation_sample']
        # anderson-darling parameter
        d_array = list()
        for i in range(len(array_x)):
            d_array.append((2 * (i + 1) - 1) * np.log(
                NormalDistribution().cdf(array_x[i], mean_array, standard_deviation))
                + (2 * (len(array_x) - (
                    i + 1)) + 1) * np.log(1 -
                                          NormalDistribution().cdf(
                                              array_x[i], mean_array,
                                              standard_deviation)))
        a_2 = - len(array_x) - 1 / \
            len(array_x) * sum(d_array)
        # anderson-darling modified
        a_m_2 = a_2 * (1 + 0.75 / len(array_x) +
                       2.25 / len(array_x) ** 2)
        # p-value
        if a_m_2 < 0.2:
            p_value = 1 - np.exp(-13.436 + 101.14 *
                                 a_m_2 - 223.73 * a_m_2 ** 2)
        elif a_m_2 < 0.340:
            p_value = 1 - np.exp(-8.318 + 42.796 * a_m_2 - 59.938 * a_m_2 ** 2)
        elif a_m_2 < 0.6:
            p_value = np.exp(0.9177 - 4.279 * a_m_2 - 1.38 * a_m_2 ** 2)
        elif a_m_2 > 0.6:
            p_value = np.exp(1.2937 - 5.709 * a_m_2 + 0.0186 * a_m_2 ** 2)
        return {
            'alpha': alpha,
            'p-value': p_value,
            'h0': 'nomally distributed data',
            'reject_h0': alpha > p_value
        }

    def shapiro_wilk(self, array_data, alpha=0.05):
        '''
        REFERENCES: http://www.portalaction.com.br/inferencia/64-teste-de-shapiro-wilk, 
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.shapiro.html
        DOCSTRING: TEST OF HYPOTHESIS SHAPIRO-WALK TO CHECK WHETER THE NORMAL DISTRIBUTIONS
            FITS A SAMPLE OF DATA
        INPUTS: LIST OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        array_data.sort()
        # turning list to numpy array
        array_x = np.array(array_data)
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.shapiro(array_x).pvalue,
            'h0': 'nomally distributed data',
            'reject_h0': alpha > stats.shapiro(array_x).pvalue
        }

    def d_agostinos_k2(self, array_data, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TEST OF HYPOTHESIS D'AGOSTINO'S K² TO CHECK WHETER THE NORMAL DISTRIBUTIONS
            FITS A SAMPLE OF DATA
        INPUTS: LIST OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        array_data.sort()
        # turning list to numpy array
        array_x = np.array(array_data)
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.normaltest(array_x).pvalue,
            'h0': 'nomally distributed data',
            'reject_h0': alpha > stats.normaltest(array_x).pvalue
        }


class CorrelationHT:

    def pearsons_correlation_coefficient(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: PEARSON'S CORRELATION COEFFICIENT, TESTING WHETER TWO SAMPLES HAVE A LINEAR 
            RELATIONSHIP - ASSUMPTIONS: I) INDEPENDENT AND EQUALLY DISTRIBUTED, II) NORMALLITY, 
            III) OBSERVATIONS IN EACH SAMPLE HAVE THE SAME VARIANCE - INTERPRETATION: H0: THE TWO 
            SAMPLES ARE INDEPENDENT, H1: THERE IS A DEPENDECY 
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.pearsonr(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.pearsonr(x1_array, x2_array).pvalue
        }

    def spearmans_rank_correlation(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: SPEARMAN'S RANK CORRELATION, TESTING WHETER TWO SAMPLES HAVE A MONOTONIC 
            RELATIONSHIP, WITH THE ASSUMPTIONS OF OBSERVATIONS IN EACH SAMPLE BEING INDEPENDET AND 
            IDENTICALLY DISTRIBUTED (IID), AS WELL, AS RANKABLE - INTERPRETATION: H0: THE TWO 
            SAMPLES ARE INDEPENDENT, H1: THERE IS A DEPENDENCY BETWEEN THE SAMPLES
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.spearmanr(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.spearmanr(x1_array, x2_array).pvalue
        }

    def kendalls_rank_correlation(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: KENDALL'S RANK CORRELATION, TESTING WHETER TWO SAMPLES HAVE A MONOTONIC 
            RELATIONSHIP, WITH THE ASSUMPTIONS OF OBSERVATIONS IN EACH SAMPLE BEING INDEPENDET AND 
            IDENTICALLY DISTRIBUTED (IID), AS WELL, AS RANKABLE - INTERPRETATION: H0: THE TWO 
            SAMPLES ARE INDEPENDENT, H1: THERE IS A DEPENDENCY BETWEEN THE SAMPLES
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.kendalltau(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.kendalltau(x1_array, x2_array).pvalue
        }

    def chi_squared_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: CHI-SQUARED TEST, TESTING WHETER TWO CATEGORICAL VARIABLES ARE RELATED OR 
            INDEPENDENT, WITH THE ASSUMPTIONS OF OBSERVATIONS USED IN THE CALCULATION OF THE 
            CONTINGENCY TABLE BEING INDEPENDENT, AND 25 OR MORE EXAMPLES IN EACH CELL OF THE 
            CONTINGENCY TABLE - INTERPRETATION: H0: THE TWO SAMPLES ARE INDEPENDET, H1: THERE IS A
            DEPENDENCY BETWEEN THE SAMPLES
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.kendalltau(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.kendalltau(x1_array, x2_array).pvalue
        }


class StationaryHT:

    def augmented_dickey_fuller(self, array_x, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER A TIME SERIES HAS A UNIT ROOT, E.G. HAS A TREND OR MORE GENERALLY
            IS AUTOREGRESSIVE - ASSUMPTIONS: OBSERVATIONS ARE TEMPORALLY ORDERED - INTERPRETATION: 
            - H0: A UNIT ROOT IS PRESENT (SERIES IS NON-COVARIANCE STATIONARY, SO IS A RANDOM WALK); 
            - HA: A UNIT ROOT IS NOT PRESENT (SERIES IS COVARIANCE STATIONARY)
        INPUTS: TIME SERIES DATA (A NUMPY ARRAY, FOR INSTANCE) AND ALPHA (0.05 AS DEFAULT)
        OUTPUTS: DICTIONARY WITH P-VALUE AND OR NOT WE OUGHT REJECT THE NON-STATIONARITY HYPOTHESIS
        '''
        # sort list
        array_x = array_x.sort()
        # turning list to numpy array
        array_x = np.array(array_x)
        # hypothesis test
        return {
            'statistic': sm.tsa.stattools.adfuller(array_x)[0],
            'p-value': sm.tsa.stattools.adfuller(array_x)[1],
            'lags': sm.tsa.stattools.adfuller(array_x)[2],
            'criteria': sm.tsa.stattools.adfuller(array_x)[3],
            'reject_h0': alpha > sm.tsa.stattools.adfuller(array_x)[1]
        }

    def kwiatkowski_phillips_schmidt_shin(self, array_x, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER A TIME SERIES IS TREND STATIONARY OR NOT - ASSUMPTIONS: 
            OBSERVATIONS ARE TEMPORALLY ORDERED - INTERPRETATION: H0: THE TIME SERIES IS 
            TREND-STATIONARY, H1: THE TIME SERIES IS NOT TREND-STATIONARY
        INPUTS: TIME SERIES DATA (A NUMPY ARRAY, FOR INSTANCE) AND ALPHA (0.05 AS DEFAULT)
        OUTPUTS: DICTIONARY WITH P-VALUE AND OR NOT WE OUGHT REJECT THE NON-STATIONARITY HYPOTHESIS
        '''
        # sort list
        array_x = array_x.sort()
        # turning list to numpy array
        array_x = np.array(array_x)
        # hypothesis test
        return {
            'statistic': sm.tsa.stattools.kpss(array_x)[0],
            'p-value': sm.tsa.stattools.kpss(array_x)[1],
            'lags': sm.tsa.stattools.kpss(array_x)[2],
            'criteria': sm.tsa.stattools.kpss(array_x)[3],
            'reject_h0': alpha > sm.tsa.stattools.kpss(array_x)[1]
        }


class MeansHT:

    def student_s_t_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER THE MEANS OF TWO INDEPENDENT SAMPLES ARE SIGNIFICANTLY DIFFERENT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), ARE NORMALLY DISTRIBUTED AND HAVE THE SAME VARIANCE - INTERPRETATION: H0: THE 
            MEANS OF THE SAMPLES ARE EQUAL, H1: THE MEANS OF THE SAMPLES ARE UNEQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.ttest_ind(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.ttest_ind(x1_array, x2_array).pvalue
        }

    def paired_student_s_t_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER THE MEANS OF TWO INDEPENDENT SAMPLES ARE SIGNIFICANTLY DIFFERENT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), ARE NORMALLY DISTRIBUTED AND HAVE THE SAME VARIANCE - INTERPRETATION: H0: THE 
            MEANS OF THE SAMPLES ARE EQUAL, H1: THE MEANS OF THE SAMPLES ARE UNEQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.ttest_rel(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.ttest_rel(x1_array, x2_array).pvalue
        }


class StatisticalDistributionsHT:

    def mann_whitney_u_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/nonparametric-statistical-significance-tests-in-python/
        DOCSTRING: TESTS WHETER THE DISTRIBUTIONS OF TWO PAIRED SAMPLES ARE EQUAL OR NOT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), CAN BE RANKED AND OBSERVATIONS ACCROSS EACH SAMPLE ARE PAIRED - INTERPRETATION: 
            H0: THE DISTRIBUTIONS OF BOTH SAMPLES ARE EQUAL, H1: THE DISTRIBUTIONS OF BOTH SAMPLES 
            ARE NOT EQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.mannwhitneyu(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.mannwhitneyu(x1_array, x2_array).pvalue
        }

    def wilcoxon_signed_rank_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER THE DISTRIBUTIONS OF TWO PAIRED SAMPLES ARE EQUAL OR NOT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), CAN BE RANKED AND OBSERVATIONS ACCROSS EACH SAMPLE ARE PAIRED - INTERPRETATION: 
            H0: THE DISTRIBUTIONS OF BOTH SAMPLES ARE EQUAL, H1: THE DISTRIBUTIONS OF BOTH SAMPLES 
            ARE NOT EQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.wilcoxon(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.wilcoxon(x1_array, x2_array).pvalue
        }

    def kruskal_wallis_h_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER THE DISTRIBUTIONS OF TWO PAIRED SAMPLES ARE EQUAL OR NOT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), CAN BE RANKED AND OBSERVATIONS ACCROSS EACH SAMPLE ARE PAIRED - INTERPRETATION: 
            H0: THE DISTRIBUTIONS OF BOTH SAMPLES ARE EQUAL, H1: THE DISTRIBUTIONS OF BOTH SAMPLES 
            ARE NOT EQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.kruskal(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.kruskal(x1_array, x2_array).pvalue
        }

    def friedman_test(self, x1_array, x2_array, alpha=0.05):
        '''
        REFERENCES: https://machinelearningmastery.com/statistical-hypothesis-tests-in-python-cheat-sheet/
        DOCSTRING: TESTS WHETER THE DISTRIBUTIONS OF TWO PAIRED SAMPLES ARE EQUAL OR NOT - 
            ASSUMPTIONS: OBSERVATIONS IN EACH SAMPLE ARE INDEPENDENT AND IDENTICALLY DISTRIBUTED (
            IID), CAN BE RANKED AND OBSERVATIONS ACCROSS EACH SAMPLE ARE PAIRED - INTERPRETATION: 
            H0: THE DISTRIBUTIONS OF BOTH SAMPLES ARE EQUAL, H1: THE DISTRIBUTIONS OF BOTH SAMPLES 
            ARE NOT EQUAL
        INPUTS: LISTS OF REAL NUMBERS AND ALPHA (STANDARD VALUE IS 0.05)
        OUTPUTS: DICT WITH ALPHA, P-VALUE AND WHETHER REJECT H0 OR NOT
        '''
        # sort list
        x1_array, x2_array = (x1_array.sort(), x2_array.sort())
        # turning list to numpy array
        x1_array, x2_array = (np.array(x1_array), np.array(x2_array))
        # hypothesis test
        return {
            'alpha': alpha,
            'p-value': stats.friedmanchisquare(x1_array, x2_array).pvalue,
            'reject_h0': alpha > stats.friedmanchisquare(x1_array, x2_array).pvalue
        }


class IndependenceHT:

    def pearson_chi_squared(self, array_y, float_significance=0.05):
        '''
        DOCSTRING: PEARSON'S CHI SQUARED TEST OF INDEPENDENCE
        INPUTS: ARRAY DATA
        OUTPUTS: DICTIONARY
        '''
        tup_ind_test = stats.chi2_contingency(array_y)
        return {
            'chi_squard_statistic': tup_ind_test[0],
            'p_value': tup_ind_test[1],
            'degrees_freedom': tup_ind_test[2],
            'array_expected_values': tup_ind_test[3],
            'significance': float_significance,
            'h0_hypothesis': 'independent',
            'reject_h0': float_significance > tup_ind_test[1]
        }


# array_data = [[48, 12, 33, 57], [34, 46, 42, 26]]
# print(IndependenceHT().pearson_chi_squared(array_data))



### COLLECTING UP2DATA B3 INFORMATION - ROOT URL: https://arquivos.b3.com.br/Web/Consolidated ###

# pypi.org libs
import pandas as pd
from requests import request
from zipfile import ZipFile
from datetime import datetime
from typing import Union
# private modules
from stpstone.settings._global_slots import YAML_B3
from stpstone.opening_config.setup import iniciating_logging
from stpstone.loggs.create_logs import CreateLog
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.object import HandlingObjects
from stpstone.handling_data.lists import HandlingLists
from stpstone.handling_data.folders import DirFilesManagement
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.lists import StrHandler
from stpstone.handling_data.xml import XMLFiles
from stpstone.loggs.db_logs import DBLogs
from stpstone.pool_conn.session import ReqSession
from stpstone.handling_data.pd import DealingPd
from stpstone.meta.validate_pm import ValidateAllMethodsMeta


class UP2DATAB3(metaclass=ValidateAllMethodsMeta):

    def __init__(self, dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1), 
                 req_session:Union[ReqSession, None]=None, bl_use_timer:bool=False) -> None:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        self.dt_ref = dt_ref
        self.req_session = ReqSession(bl_use_timer=bl_use_timer) \
            if req_session is None else req_session
        self.token = self.access_token
    
    @property
    def access_token(self) -> str:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        url_token = YAML_B3['up2data_b3']['request_token'].format(self.dt_ref)
        req_resp = self.req_session.session.get(url_token)
        req_resp.raise_for_status()
        return req_resp.json()[YAML_B3['up2data_b3']['key_token']]
    
    @property
    def instruments_register(self) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # requesting data
        resp_req = self.req_session.session.get(
            YAML_B3['up2data_b3']['access_api'].format(self.token))
        resp_req.raise_for_status()
        str_resp_req = resp_req.text
        # cleaning response
        list_resp_req = str_resp_req.split('\n')
        list_headers = list_resp_req[1].split(';')
        list_data = list_resp_req[2:]
        for row in list_data:
            if len(row) == 0: continue
            list_row = row.split(';')
            list_ser.append(dict(zip(list_headers, list_row)))
        # dataframe from serialized list
        df_ = pd.DataFrame(list_ser)
        # changing columns case, from cammel to snake
        df_.columns = [StrHandler().camel_to_snake(x) for x in df_.columns]
        # changing column types
        print('cols date: ', YAML_B3['instruments_register_row']['cols_date'])
        df_ = DealingPd().pipeline_df_startup(
            df_,
            YAML_B3['instruments_register_row']['cols_dtype'],
            YAML_B3['instruments_register_row']['cols_date'],
            str_fmt_dt=YAML_B3['instruments_register_row']['str_fmt_dt']
        )
        # adding logging
        df_ = DBLogs().audit_log(
            df_, 
            DatesBR().utc_from_dt(DatesBR().curr_date), 
            DatesBR().utc_log_ts
        )
        # returning instruments register
        return df_
    
    @property
    def daily_trades_secondary_market(self) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # requesting daily trades
        file_daily_trades = DirFilesManagement().get_zip_from_web_in_memory(
            YAML_B3['up2data_b3']['url_daily_trades'].format(self.dt_ref.strftime('%Y-%m-%d')), 
            bl_io_interpreting=YAML_B3['up2data_b3']['bl_io_interpreting'],
            bl_verify=YAML_B3['up2data_b3']['bl_verify']
        )
        # importing to pandas dataframe
        reader = pd.read_csv(
            file_daily_trades,
            sep=YAML_B3['up2data_b3']['str_sep'], 
            decimal=YAML_B3['up2data_b3']['str_decimal']
        )
        df_daily_trades = pd.DataFrame(reader)
        # substituindo valores nulos
        df_daily_trades.fillna(YAML_B3['up2data_b3']['value_nan'], inplace=True)
        # changing columns types
        df_daily_trades = DealingPd().change_dtypes(
            df_daily_trades, 
            {
                YAML_B3['up2data_b3']['col_dt_ref']: str,
                YAML_B3['up2data_b3']['col_instrument']: str,
                YAML_B3['up2data_b3']['col_stock_update']: int,
                YAML_B3['up2data_b3']['col_pv']: float,
                YAML_B3['up2data_b3']['col_qty_trd']: int,
                YAML_B3['up2data_b3']['col_closing_time']: int,
                YAML_B3['up2data_b3']['col_trd_code']: int,
                YAML_B3['up2data_b3']['col_section_type']: int,
                YAML_B3['up2data_b3']['col_trd_dt']: str,
                YAML_B3['up2data_b3']['col_buyer_code']: int,
                YAML_B3['up2data_b3']['col_seller_code']: int
            }, 
            [YAML_B3['up2data_b3']['col_dt_ref'], YAML_B3['up2data_b3']['col_trd_dt']], 
            errors='ignore'
        )
        # column asset
        df_daily_trades[YAML_B3['up2data_b3']['col_create_asset']] = [
            row[YAML_B3['up2data_b3']['col_instrument']][:3] 
            for _, row in df_daily_trades.iterrows()
        ]
        # adding logging
        df_daily_trades = DBLogs().audit_log(
            df_daily_trades, 
            DatesBR().utc_from_dt(self.dt_ref), 
            DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_daily_trades

    @property
    def security_category_name(self, key_ticker='TckrSymb',
                               col_security_category_name='SctyCtgyNm',
                               col_market_name='MktNm',
                               bl_return_markets_not_classified=False):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_markets_classified = list()
        dict_export = dict() 
        # fetch in memory instruments register of assets traded in b3 exchange
        df_ = self.instruments_register_raw
        print('*** REIGSTER B3 RAW ***')
        print(df_)
        # creating dictionary with instruments according to each type of market
        for security_division, col_ in [
            ('securities_by_category_name', col_security_category_name),
                ('securities_by_market_name', col_market_name)]:
            for market, list_source_names in YAML_B3['up2data_b3'][security_division].items():
                #   validating wheter the market is already in the exporting dictionary and extending
                #       or creating a new list
                if market not in dict_export:
                    dict_export[market] = HandlingLists().remove_duplicates(
                        df_[
                            df_[col_].isin(
                                list_source_names)][key_ticker].tolist())
                else:
                    dict_export[market].extend(HandlingLists().remove_duplicates(
                        df_[
                            df_[col_].isin(
                                list_source_names)][key_ticker].tolist()))
                list_markets_classified.extend(list_source_names)
        # defining markets not classified, if is user's will
        if bl_return_markets_not_classified == True:
            return HandlingLists().remove_duplicates(
                df_[
                    ~df_[col_security_category_name].isin(
                        list_markets_classified)][col_security_category_name].tolist())
        else:
            #   returing dictionary with tickers accordingly to the market type which participates
            return dict_export

    def loan_balace(self, wd_bef=1, url='https://arquivos.b3.com.br/tabelas/table/LoanBalance/{}/{}', 
                    method='GET', key_pg_count='pageCount', key_valores='values', 
                    list_cols_lb_b3=['RptDt', 'TckrSymb', 'Asst', 'QtyCtrctsDay', 'QtyShrDay', 
                                     'DnrMinRate', 'DnrAvrgRate', 'DnrMaxRate', 'TakrMinRate', 
                                     'TakrAvrgRate', 'TakrMaxRate', 'MktNm']):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # reference date
        data_ref_lb_b3 = DatesBR().sub_working_days_before(
            DatesBR().curr_date, wd_bef).strftime('%Y-%m-%d')
        # payload
        dict_payload = {}
        # headers
        dict_headers = {
            'authority': 'arquivos.b3.com.br',
            'accept': '*/*',
            'accept-language': 'pt-BR,pt;q=0.9,en-US;q=0.8,en;q=0.7',
            'cookie': 'OptanonAlertBoxClosed=2023-03-24T13:33:20.482Z; _ga_T6D1M7G116=GS1.1.1684332031.17.0.1684332031.60.0.0; _tt_enable_cookie=1; _ttp=-3ewIEHyq8lWLAqJVQ6O6T5fGqv; _ga_X5KRPBP7ZE=GS1.3.1690829338.5.1.1690829690.60.0.0; _ga_0W7NXV5699=GS1.1.1695259533.1.0.1695259540.0.0.0; lumClientId=FF8080818A6B1B5C018ABCE5D291602D; lumMonUid=l_QNKA5sbq-l9hkDYeDpkuHl5Qvn8pSA; _ga_F2HQZ24Y48=GS1.1.1695386295.1.1.1695386413.36.0.0; visid_incap_2246223=zCpOWadTSVqdEX/oNXZmLb0vl2UAAAAAQUIPAAAAAAArXUd3AyxcF+G4fbInAZ/+; _gcl_au=1.1.1328610258.1704896955; _gcl_aw=GCL.1709150038.CjwKCAiA0PuuBhBsEiwAS7fsNTm-phWyGfd21MQsHApfiDnWpzvVnpH-JCKFzGOxLwGYnYogLKSlFBoCQXQQAvD_BwE; _ga_5E2DT9ZLVR=GS1.1.1709150038.3.1.1709150112.60.0.0; _ga_CNJN5WQC5G=GS1.1.1709239346.20.0.1709239476.0.0.0; _ga_CRYYSCYF12=GS1.1.1709825617.2.1.1709825617.60.0.0; _ga_FTT9L7SR7B=GS1.1.1709827350.1.1.1709829244.54.0.0; nlbi_2246223=LPzcM1dvj3Qj9PJY9OkOmwAAAAAHcT0kwIANYP4of6jRlfTo; incap_ses_1239_2246223=V2rbSa8xMQgKwd16LtExEbuC+WUAAAAA6Bx9qzgrW2qvO8fGb1s7Ug==; dtCookie=v_4_srv_33_sn_B96EE88FC773C73EDD967E7FE8D41023_perc_100000_ol_0_mul_1_app-3Afd69ce40c52bd20e_0_rcs-3Acss_0; _gid=GA1.3.971339064.1710850908; _clck=v6volp%7C2%7Cfk7%7C0%7C1232; auth0=; TS0171d45d=011d592ce15f7772eec4672b5224a65a3ed9ad68450822231fe43032d6783dbc2d826bb431969fefc876b99ec27a5a6eacbd316945; _ga_SS7FXRTPP3=GS1.1.1710850908.138.1.1710852407.59.0.0; _ga=GA1.3.154973072.1679664794; OptanonConsent=isGpcEnabled=0&datestamp=Tue+Mar+19+2024+09%3A46%3A47+GMT-0300+(Hor%C3%A1rio+Padr%C3%A3o+de+Bras%C3%ADlia)&version=6.21.0&isIABGlobal=false&hosts=&landingPath=NotLandingPage&groups=C0003%3A1%2CC0001%3A1%2CC0004%3A1%2CC0002%3A1&geolocation=%3B&AwaitingReconsent=false; _clsk=hrfzvd%7C1710852408467%7C8%7C1%7Cb.clarity.ms%2Fcollect; incap_ses_1239_2246223=9YpWOoNjqRP6zeF6LtExEfGI+WUAAAAASCaRRjwRJW3uB4oCgh0+dg==; nlbi_2246223=fL1wUKg2iFtTP+KU9OkOmwAAAABL1G2EaiPMxRnkbfBf01Ql; visid_incap_2246223=WX+mv0j8SJ2ikziYIwjLooCG+WUAAAAAQUIPAAAAAAD1iQw5sjPBQa6Xku7lrpeD; auth0=',
            'sec-ch-ua': '"Chromium";v="122", "Not(A:Brand";v="24", "Google Chrome";v="122"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36',
            'x-elastica_gw': '3.116.0'
        }
        # retrieving loan balance b3
        resp_req = request(method, url.format(data_ref_lb_b3, 1), headers=dict_headers, 
                           data=dict_payload)
        # raise error when status code is different from 2xx
        resp_req.raise_for_status()
        # storying current json in memory
        json_lb_b3 = resp_req.json()
        # verificando o número de páginas a serem consultadas
        int_pgs_lb_b3 = json_lb_b3[key_pg_count]
        # loopando em torno das páginas de interesse
        for i in range(1, int_pgs_lb_b3 + 1):
            #   retrieving loan balance b3
            resp_req = request(method, url.format(data_ref_lb_b3, i), headers=dict_headers, 
                               data=dict_payload)
            #   raise error when status code is different from 2xx
            resp_req.raise_for_status()
            #   storying current json in memory
            json_lb_b3 = resp_req.json()
            #   looping within available values
            for list_ in json_lb_b3[key_valores]:
                #   appending to serialized list
                list_ser.append(
                    dict(zip(list_cols_lb_b3, list_))
                )
        # builiding dataframe
        df_lb_b3 = pd.DataFrame(list_ser)
        # dropping duplicates
        df_lb_b3.drop_duplicates(inplace=True)
        # changing column types
        df_lb_b3 = df_lb_b3.astype({
            list_cols_lb_b3[0]: str,
            list_cols_lb_b3[1]: str,
            list_cols_lb_b3[2]: str,
            list_cols_lb_b3[3]: float,
            list_cols_lb_b3[4]: float,
            list_cols_lb_b3[5]: float,
            list_cols_lb_b3[6]: float,
            list_cols_lb_b3[7]: float,
            list_cols_lb_b3[8]: float,
            list_cols_lb_b3[9]: float,
            list_cols_lb_b3[10]: float,
            list_cols_lb_b3[11]: str
        })
        df_lb_b3[list_cols_lb_b3[0]] = [DatesBR().timestamp_separator_string_to_datetime(d) 
                                        for d in df_lb_b3[list_cols_lb_b3[0]]]
        # sorting dataframe
        df_lb_b3.sort_values([
            list_cols_lb_b3[0],
            list_cols_lb_b3[1],
            list_cols_lb_b3[2],
            list_cols_lb_b3[-1]
        ], ascending=[False, True, True, True], inplace=True)
        # adding logging
        df_lb_b3 = DBLogs().audit_log(
            df_lb_b3, 
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, wd_bef)
            ), 
            DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_lb_b3

    def lending_open_position(self, wd_bef=1, url='https://arquivos.b3.com.br/tabelas/table/LendingOpenPosition/{}/{}', 
                              method='GET', key_pg_count='pageCount', key_valores='values', 
                              list_cols_lop_b3=['RptDt', 'TckrSymb', 'Asst', 'BalQty', 'TradAvrgPric', 'PricFctr', 'BalVal']):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # reference date
        data_ref_lb_b3 = DatesBR().sub_working_days_before(
            DatesBR().curr_date, wd_bef).strftime('%Y-%m-%d')
        # payload
        dict_payload = {}
        # headers
        dict_headers = {
            'authority': 'arquivos.b3.com.br',
            'accept': '*/*',
            'accept-language': 'pt-BR,pt;q=0.9,en-US;q=0.8,en;q=0.7',
            'cookie': 'OptanonAlertBoxClosed=2023-03-24T13:33:20.482Z; _ga_T6D1M7G116=GS1.1.1684332031.17.0.1684332031.60.0.0; _tt_enable_cookie=1; _ttp=-3ewIEHyq8lWLAqJVQ6O6T5fGqv; _ga_X5KRPBP7ZE=GS1.3.1690829338.5.1.1690829690.60.0.0; _ga_0W7NXV5699=GS1.1.1695259533.1.0.1695259540.0.0.0; lumClientId=FF8080818A6B1B5C018ABCE5D291602D; lumMonUid=l_QNKA5sbq-l9hkDYeDpkuHl5Qvn8pSA; _ga_F2HQZ24Y48=GS1.1.1695386295.1.1.1695386413.36.0.0; visid_incap_2246223=zCpOWadTSVqdEX/oNXZmLb0vl2UAAAAAQUIPAAAAAAArXUd3AyxcF+G4fbInAZ/+; _gcl_au=1.1.1328610258.1704896955; _gcl_aw=GCL.1709150038.CjwKCAiA0PuuBhBsEiwAS7fsNTm-phWyGfd21MQsHApfiDnWpzvVnpH-JCKFzGOxLwGYnYogLKSlFBoCQXQQAvD_BwE; _ga_5E2DT9ZLVR=GS1.1.1709150038.3.1.1709150112.60.0.0; _ga_CNJN5WQC5G=GS1.1.1709239346.20.0.1709239476.0.0.0; _ga_CRYYSCYF12=GS1.1.1709825617.2.1.1709825617.60.0.0; _ga_FTT9L7SR7B=GS1.1.1709827350.1.1.1709829244.54.0.0; nlbi_2246223=LPzcM1dvj3Qj9PJY9OkOmwAAAAAHcT0kwIANYP4of6jRlfTo; incap_ses_1239_2246223=V2rbSa8xMQgKwd16LtExEbuC+WUAAAAA6Bx9qzgrW2qvO8fGb1s7Ug==; dtCookie=v_4_srv_33_sn_B96EE88FC773C73EDD967E7FE8D41023_perc_100000_ol_0_mul_1_app-3Afd69ce40c52bd20e_0_rcs-3Acss_0; _gid=GA1.3.971339064.1710850908; _clck=v6volp%7C2%7Cfk7%7C0%7C1232; auth0=; TS0171d45d=011d592ce15f7772eec4672b5224a65a3ed9ad68450822231fe43032d6783dbc2d826bb431969fefc876b99ec27a5a6eacbd316945; _ga_SS7FXRTPP3=GS1.1.1710850908.138.1.1710852407.59.0.0; _ga=GA1.3.154973072.1679664794; OptanonConsent=isGpcEnabled=0&datestamp=Tue+Mar+19+2024+09%3A46%3A47+GMT-0300+(Hor%C3%A1rio+Padr%C3%A3o+de+Bras%C3%ADlia)&version=6.21.0&isIABGlobal=false&hosts=&landingPath=NotLandingPage&groups=C0003%3A1%2CC0001%3A1%2CC0004%3A1%2CC0002%3A1&geolocation=%3B&AwaitingReconsent=false; _clsk=hrfzvd%7C1710852408467%7C8%7C1%7Cb.clarity.ms%2Fcollect; incap_ses_1239_2246223=9YpWOoNjqRP6zeF6LtExEfGI+WUAAAAASCaRRjwRJW3uB4oCgh0+dg==; nlbi_2246223=fL1wUKg2iFtTP+KU9OkOmwAAAABL1G2EaiPMxRnkbfBf01Ql; visid_incap_2246223=WX+mv0j8SJ2ikziYIwjLooCG+WUAAAAAQUIPAAAAAAD1iQw5sjPBQa6Xku7lrpeD; auth0=',
            'sec-ch-ua': '"Chromium";v="122", "Not(A:Brand";v="24", "Google Chrome";v="122"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36',
            'x-elastica_gw': '3.116.0'
        }
        # retrieving loan balance b3
        resp_req = request(method, url.format(data_ref_lb_b3, 1), headers=dict_headers, 
                           data=dict_payload)
        # raise error when status code is different from 2xx
        resp_req.raise_for_status()
        # storying current json in memory
        json_lb_b3 = resp_req.json()
        # verificando o número de páginas a serem consultadas
        int_pgs_lb_b3 = json_lb_b3[key_pg_count]
        # loopando em torno das páginas de interesse
        for i in range(1, int_pgs_lb_b3 + 1):
            #   retrieving loan balance b3
            resp_req = request(method, url.format(data_ref_lb_b3, i), headers=dict_headers, 
                               data=dict_payload)
            #   raise error when status code is different from 2xx
            resp_req.raise_for_status()
            #   storying current json in memory
            json_lb_b3 = resp_req.json()
            #   looping within available values
            for list_ in json_lb_b3[key_valores]:
                #   appending to serialized list
                list_ser.append(
                    dict(zip(list_cols_lop_b3, list_))
                )
        # builiding dataframe
        df_lop_b3 = pd.DataFrame(list_ser)
        # dropping duplicates
        df_lop_b3.drop_duplicates(inplace=True)
        # changing column types
        df_lop_b3 = df_lop_b3.astype({
            list_cols_lop_b3[0]: str,
            list_cols_lop_b3[1]: str,
            list_cols_lop_b3[2]: str,
            list_cols_lop_b3[3]: float,
            list_cols_lop_b3[4]: float,
            list_cols_lop_b3[5]: float,
            list_cols_lop_b3[6]: float
        })
        df_lop_b3[list_cols_lop_b3[0]] = [DatesBR().timestamp_separator_string_to_datetime(d) 
                                        for d in df_lop_b3[list_cols_lop_b3[0]]]
        # sorting dataframe
        df_lop_b3.sort_values([
            list_cols_lop_b3[0],
            list_cols_lop_b3[1],
            list_cols_lop_b3[2],
            list_cols_lop_b3[-1]
        ], ascending=[False, True, True, True], inplace=True)
        # adding logging
        df_lop_b3 = DBLogs().audit_log(
            df_lop_b3, 
            DatesBR().utc_from_dt(
                DatesBR().sub_working_days(DatesBR().curr_date, wd_bef)
            ), 
            DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_lop_b3
    
    @property
    def bond_issuers_accepted_warranty_b3(self, url_banks_brazil='http://www.bcb.gov.br/pom/spb/estatistica/port/ParticipantesSTRport.csv', 
        url_bond_issuers_warranty_b3='https://www.b3.com.br/data/files/86/34/44/30/407E971089F29C97AC094EA8/Emissores%20Aceitos%20-%20Titulos%20Privados.zip', 
        bl_io_interpreting=False, col_issuers='ISSUERS', col_ispb='ISPB', col_numero_cod='NUMERO_CODIGO', 
        col_nome_extenso='NOME_EXTENSO', col_fr_total='FRTotal',
        dict_replace_str_name= {
            'S/A': '',
            'S.A.': '', 
            'S.A': '',
            ' SA': '',
            'LTDA.': '',
            'LTDA': '',
            '(BRASIL)': '', 
            'Brasil S.A.': '',
            'BRASIL S A': '',
            'DE INVESTIMENTO': '', 
            'DE INVEST. ': '', 
            'BCO': 'BANCO', 
            'J.P.MORGAN': 'J.P. MORGAN', 
            'CHINA BRASIL': 'CHINA'
        }):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # importing to memory banks currently operating domestically
        reader = pd.read_csv(url_banks_brazil)
        df_banks_operating_domestically = pd.DataFrame(reader)
        # change column names
        df_banks_operating_domestically.rename(columns=dict(zip(
            df_banks_operating_domestically.columns, 
            [StrHandler().remove_diacritics(str(x).upper()) for x in 
                df_banks_operating_domestically.columns])), inplace=True)
        # defining list of columns
        list_cols = df_banks_operating_domestically.columns
        # change columns types
        df_banks_operating_domestically.fillna(0, inplace=True)
        df_banks_operating_domestically = df_banks_operating_domestically.astype({
            col_ispb: int,
            col_numero_cod: int
        })
        df_banks_operating_domestically = df_banks_operating_domestically.astype({
            col_ispb: str,
            col_numero_cod: str,
            col_nome_extenso: str, 
            col_fr_total: float
        })
        # dealing with complete names from financial institutions
        df_banks_operating_domestically[col_nome_extenso] = [StrHandler().remove_diacritics(
            StrHandler().replace_all(str(x).upper(), dict_replace_str_name).strip()) 
            for x in df_banks_operating_domestically[col_nome_extenso].tolist()]
        # importing to memory banks accepted as warranty in b3 exchange stock market
        list_unziped_files = DirFilesManagement().get_zip_from_web_in_memory(
            url_bond_issuers_warranty_b3, 
            bl_io_interpreting=bl_io_interpreting)
        if type(list_unziped_files) != list:
            list_unziped_files = [list_unziped_files]
        # looping through files in .zip and uploading to dataframe
        for nome_arquivo_xlsx in list_unziped_files:
            #   importing to dataframe
            reader = pd.read_excel(nome_arquivo_xlsx, encoding='utf-8')
            df_ = pd.DataFrame(reader)
            #   changing format to list of dictionaries
            list_banks_warranty_b3 = df_[list(df_.columns)[0]].tolist()
        # removing duplicates
        list_banks_warranty_b3 = HandlingLists().remove_duplicates(
            list_banks_warranty_b3)
        list_banks_warranty_b3 = [str(x).upper() for x in list_banks_warranty_b3]
        # creating dataframe with banks accepted as warranty in b3 clearing
        df_banks_accepted_warranty_b3 = pd.DataFrame([{col_issuers: bank} for bank in 
            list_banks_warranty_b3])
        df_banks_accepted_warranty_b3.sort_values([col_issuers], ascending=[True], inplace=True)
        # dealing with complete names from financial institutions
        df_banks_accepted_warranty_b3[col_issuers] = [StrHandler().remove_diacritics(
            StrHandler().replace_all(str(x).upper(), dict_replace_str_name).strip()) 
            for x in df_banks_accepted_warranty_b3[col_issuers].tolist()]
        # limiting banks accepted as warranty in b3 clearing
        df_banks_acc_warr = df_banks_operating_domestically[df_banks_operating_domestically[
            col_nome_extenso].isin(df_banks_accepted_warranty_b3[col_issuers])]
        # limitating columns of interest
        df_banks_acc_warr = df_banks_acc_warr[list_cols]
        # adding logging
        df_ = DBLogs().audit_log(
            df_, 
            DatesBR().utc_from_dt(DatesBR().curr_date), 
            DatesBR().utc_log_ts
        )
        # returning dataframe of inner join
        return df_banks_acc_warr



import typing
from .client import Client


class Responses:
    """Typeform Responses API client"""

    def __init__(self, client: Client):
        """Constructor for Typeform Responses class"""
        self.__client = client

    def list(
        self, uid: str, pageSize: int = None, since: str = None, until: str = None,
        after: str = None, before: str = None, includedResponseIds: str = None,
        completed: bool = None, sort: str = None, query: str = None, fields: typing.List[str] = None
    ) -> dict:
        """
        Returns form responses and date and time of form landing and submission.
        """
        return self.__client.request('get', '/forms/%s/responses' % uid, params={
            'page_size': pageSize or None,
            'since': since or None,
            'until': until,
            'after': after,
            'before': before,
            'included_response_ids': includedResponseIds,
            'completed': completed,
            'sort': sort,
            'query': query,
            'fields': fields
        })

    def delete(self, uid: str, includedTokens: typing.Union[str, typing.List[str]]) -> str:
        """
        Delete responses to a form. You must specify the `included_tokens`/`includedTokens` parameter.
        Return a `str` based on success of deletion, `OK` on success, otherwise an error message.
        """
        return self.__client.request('delete', '/forms/%s/responses' % uid, params={
            'included_tokens': includedTokens
        })



### CLASS B-TREE ###

from typing import List, Optional, Tuple


class Node:

    def __init__(self, bl_leaf:bool = False) -> None:
        '''
        DOCSTRING: INITIALIZE A NODE FOR THE B-TREE
        INPUT: LEAF (BOOL): INDICATES IF THE NODE IS A LEAF
        OUTPUT: NONE
        '''
        self.keys:List[int] = []
        self.children:List['Node'] = []
        self.bl_leaf:bool = bl_leaf


class BTree:

    def __init__(self, int_t:int) -> None:
        '''
        REFERENCES:
            https://www.geeksforgeeks.org/b-tree-in-python/
            https://github.com/msambol/dsa/blob/master/trees/b_tree.py
            https://www.youtube.com/michaelsambol
        DOCSTRING: INITIALIZE A B-TREE WITH A GIVEN MINIMUM DEGREE
        INPUT: T (INT): MINIMUM DEGREE OF THE B-TREE (DEFINES THE RANGE OF CHILDREN FOR EACH NODE)
        OUTPUT: NONE
        '''
        self.root:Node = Node(True)
        self.int_t:int = int_t

    def search(self, int_key:int, node:Optional[Node]=None) -> Optional[Tuple[Node, int]]:
        '''
        DOCSTRING: SEARCH FOR A KEY IN THE B-TREE STARTING FROM A GIVEN NODE
        INPUT:
            - KEY (INT): THE KEY TO SEARCH FOR
            - NODE (OPTIONAL[NODE]): THE NODE TO START THE SEARCH FROM (DEFAULT IS THE ROOT)
        OUTPUT:
            OPTIONAL[TUPLE[NODE, INT]]: THE NODE AND INDEX OF THE KEY IF FOUND, OTHERWISE NONE
        '''
        node = self.root if node is None else node
        int_i = 0
        while int_i < len(node.keys) and int_key > node.keys[int_i]:
            int_i += 1
        if int_i < len(node.keys) and int_key == node.keys[int_i]:
            return node, int_i
        elif node.bl_leaf:
            return None
        else:
            return self.search(int_key, node.children[int_i])

    def split_child(self, node_x:Node, int_i:int) -> None:
        '''
        DOCSTRING: SPLIT A FULL CHILD NODE INTO TWO AND ADJUST THE PARENT NODE
        INPUT:
            - X (NODE): THE PARENT NODE
            - I (INT): THE INDEX OF THE CHILD TO SPLIT
        OUTPUT: NONE
        '''
        # y is a full child of node_x
        node_y = node_x.children[int_i]
        # create a new node and add it to node_x's list of children
        node_z = Node(node_y.bl_leaf)
        node_x.children.insert(int_i + 1, node_z)
        # insert the median of the full child y into node_x
        node_x.keys.insert(int_i, node_y.keys[self.int_t - 1])
        # split apart y's keys into y & z
        node_z.keys = node_y.keys[self.int_t:(2 * self.int_t) - 1]
        node_y.keys = node_y.keys[0:self.int_t - 1]
        # if y is not a leaf, we reassign y's children to y & z
        if not node_y.bl_leaf:
            node_z.children = node_y.children[self.int_t:2 * self.int_t]
            node_y.children = node_y.children[0:self.int_t]

    def insert(self, int_k:int) -> None:
        '''
        DOCSTRING: INSERT A KEY INTO THE B-TREE
        INPUT:
            - K (INT): THE KEY TO INSERT
        OUTPUT: NONE
        '''
        # if root is full, create a new node - tree's height grows by 1
        if len(self.root.keys) == (2 * self.int_t) - 1:
            new_root = Node()
            self.root = new_root
            new_root.children.insert(0, self.root)
            self.split_child(new_root, 0)
            self.insert_non_full(new_root, int_k)
        else:
            self.insert_non_full(self.root, int_k)

    def insert_non_full(self, node_x:Node, int_k:int) -> None:
        '''
        DOCSTRING: INSERT A KEY INTO A NODE THAT IS NOT FULL.
        INPUT:
            X (NODE): THE NODE TO INSERT THE KEY INTO.
            K (INT): THE KEY TO INSERT.
        OUTPUT:
            NONE
        '''
        int_i = len(node_x.keys) - 1
        # find the correct spot in the leaf to insert the key
        if node_x.bl_leaf:
            #   placeholder
            node_x.keys.append(None)
            while int_i >= 0 and int_k < node_x.keys[int_i]:
                node_x.keys[int_i + 1] = node_x.keys[int_i]
                int_i -= 1
            node_x.keys[int_i + 1] = int_k
        # if not a leaf, find the correct subtree to insert the key
        else:
            while int_i >= 0 and int_k < node_x.keys[int_i]:
                int_i -= 1
            int_i += 1
            #   if child node is full, split it
            if len(node_x.children[int_i].keys) == (2 * self.int_t) - 1:
                self.split_child(node_x, int_i)
                if int_k > node_x.keys[int_i]:
                    int_i += 1
            self.insert_non_full(node_x.children[int_i], int_k)

    def delete(self, node_x:Node, int_k:int) -> None:
        '''
        DOCSTRING: DELETE A KEY FROM THE B-TREE STARTING FROM A GIVEN NODE.
        INPUT:
            - X (NODE): THE NODE TO START THE DELETION FROM
            - K (INT): THE KEY TO DELETE
        OUTPUT:
            NONE
        '''
        int_i = 0
        while int_i < len(node_x.keys) and int_k > node_x.keys[int_i]:
            int_i += 1
        if node_x.leaf:
            if int_i < len(node_x.keys) and node_x.keys[int_i] == int_k:
                node_x.keys.pop(int_i)
            return
        if int_i < len(node_x.keys) and node_x.keys[int_i] == int_k:
            return self.delete_internal_node(node_x, int_k, int_i)
        elif len(node_x.children[int_i].keys) >= self.self.t:
            self.delete(node_x.children[int_i], int_k)
        else:
            if int_i != 0 and int_i + 2 < len(node_x.children):
                if len(node_x.children[int_i - 1].keys) >= self.self.t:
                    self.delete_sibling(node_x, int_i, int_i - 1)
                elif len(node_x.children[int_i + 1].keys) >= self.self.t:
                    self.delete_sibling(node_x, int_i, int_i + 1)
                else:
                    self.delete_merge(node_x, int_i, int_i + 1)
            elif int_i == 0:
                if len(node_x.children[int_i + 1].keys) >= self.self.t:
                    self.delete_sibling(node_x, int_i, int_i + 1)
                else:
                    self.delete_merge(node_x, int_i, int_i + 1)
            elif int_i + 1 == len(node_x.children):
                if len(node_x.children[int_i - 1].keys) >= self.self.t:
                    self.delete_sibling(node_x, int_i, int_i - 1)
                else:
                    self.delete_merge(node_x, int_i, int_i - 1)
            self.delete(node_x.children[int_i], int_k)

    def delete_internal_node(self, node_x:Node, node_k:None, int_i:int) -> None:
        '''
        DOCSTRING: DELETE INTERNAL NODE
        INPUTS:
            - X (NODE): THE NODE TO START THE DELETION FROM
            - K (INT): THE KEY TO DELETE
            - I (INT): THE INDEX OF THE KEY TO DELETE
        OUTPUTS:
        '''
        if node_x.leaf:
            if node_x.keys[int_i] == node_k:
                node_x.keys.pop(int_i)
            return

        if len(node_x.children[int_i].keys) >= self.t:
            node_x.keys[int_i] = self.delete_predecessor(node_x.children[int_i])
            return
        elif len(node_x.children[int_i + 1].keys) >= self.t:
            node_x.keys[int_i] = self.delete_successor(node_x.children[int_i + 1])
            return
        else:
            self.delete_merge(node_x, int_i, int_i + 1)
            self.delete_internal_node(node_x.children[int_i], node_k, self.self.t - 1)

    def delete_predecessor(self, node_x:None) -> None:
        '''
        DOCSTRING: DELETE PREDECESSOR
        INPUTS:
            - X (NODE): THE NODE TO START THE DELETION FROM
        OUTPUTS: NONE
        '''
        if node_x.leaf:
            return node_x.keys.pop()
        n = len(node_x.keys) - 1
        if len(node_x.children[n].keys) >= self.self.t:
            self.delete_sibling(node_x, n + 1, n)
        else:
            self.delete_merge(node_x, n, n + 1)
        self.delete_predecessor(node_x.children[n])

    def delete_successor(self, node_x:Node) -> None:
        '''
        DOCSTRING: DELETE SUCCESSOR
        INPUTS:
            - X (NODE): THE NODE TO START THE DELETION FROM
        OUTPUTS: NONE
        '''
        if node_x.leaf:
            return node_x.keys.pop(0)
        if len(node_x.children[1].keys) >= self.self.t:
            self.delete_sibling(node_x, 0, 1)
        else:
            self.delete_merge(node_x, 0, 1)
        self.delete_successor(node_x.children[0])

    def delete_merge(self, node_x:Node, int_i:int, int_j:int) -> None:
        '''
        DOCSTRING: MERGE TWO NODES
        INPUTS:
            - X (NODE): THE NODE TO START THE MERGE FROM
            - I (INT): THE INDEX OF THE NODE TO MERGE WITH
            - J (INT): THE INDEX OF THE NODE TO MERGE WITH
        OUTPUTS: NONE
        '''
        cnode = node_x.children[int_i]
        if int_j > int_i:
            rsnode = node_x.children[int_j]
            cnode.keys.append(node_x.keys[int_i])
            for node_k in range(len(rsnode.keys)):
                cnode.keys.append(rsnode.keys[node_k])
                if len(rsnode.children) > 0:
                    cnode.children.append(rsnode.children[node_k])
            if len(rsnode.children) > 0:
                cnode.children.append(rsnode.children.pop())
            new = cnode
            node_x.keys.pop(int_i)
            node_x.children.pop(int_j)
        else:
            lsnode = node_x.children[int_j]
            lsnode.keys.append(node_x.keys[int_j])
            for int_i in range(len(cnode.keys)):
                lsnode.keys.append(cnode.keys[int_i])
                if len(lsnode.children) > 0:
                    lsnode.children.append(cnode.children[int_i])
            if len(lsnode.children) > 0:
                lsnode.children.append(cnode.children.pop())
            new = lsnode
            node_x.keys.pop(int_j)
            node_x.children.pop(int_i)
        if node_x == self.root and len(node_x.keys) == 0:
            self.root = new

    def delete_sibling(self, node_x:Node, int_i:int, int_j:int) -> None:
        '''
        DOCSTRING: DELETE A SIBLING NODE
        INPUTS:
            - X (NODE): THE NODE TO START THE DELETION FROM
            - I (INT): THE INDEX OF THE KEY TO DELETE
            - J (INT): THE INDEX OF THE SIBLING NODE TO MERGE WITH
        OUTPUTS: NONE
        '''
        cnode = node_x.children[int_i]
        if int_i < int_j:
            rsnode = node_x.children[int_j]
            cnode.keys.append(node_x.keys[int_i])
            node_x.keys[int_i] = rsnode.keys[0]
            if len(rsnode.children) > 0:
                cnode.children.append(rsnode.children[0])
                rsnode.children.pop(0)
            rsnode.keys.pop(0)
        else:
            lsnode = node_x.children[int_j]
            cnode.keys.insert(0, node_x.keys[int_i - 1])
            node_x.keys[int_i - 1] = lsnode.keys.pop()
            if len(lsnode.children) > 0:
                cnode.children.insert(0, lsnode.children.pop())

    def print_tree(self, node_x: Node, int_level:int = 0) -> None:
        '''
        DOCSTRING: PRINT THE STRUCTURE OF THE B-TREE.
        INPUT:
            X (NODE): THE NODE TO START PRINTING FROM.
            LEVEL (INT): THE CURRENT LEVEL IN THE TREE.
        OUTPUT:
            NONE
        '''
        print(f'Level {int_level}', end=": ")
        for int_i in node_x.keys:
            print(int_i, end=" ")
        print()
        int_level += 1
        if len(node_x.children) > 0:
            for child in node_x.children:
                self.print_tree(child, int_level)


from unittest import TestCase
import requests_mock
import urllib.parse

from .fixtures import TOKEN

from typeform import Typeform
from typeform.constants import API_BASE_URL


class FormsTestCase(TestCase):
    def setUp(self):
        self.forms = Typeform(TOKEN).forms
        form = self.forms.create({
            'title': 'title'
        })
        self.formID = form.get('id')

    def tearDown(self):
        list = self.forms.list()
        forms = list.get('items', [])
        for form in forms:
            self.forms.delete(form.get('id'))

    def test_forms_returns_method_and_path(self):
        """
        get all forms has the correct method and path
        """
        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms', json={})
            self.forms.list()

            history = m.request_history
            self.assertEqual(history[0].url, API_BASE_URL+'/forms')
            self.assertEqual(history[0].method, 'GET')

    def test_forms_correct_params(self):
        """
        paramters are sent correctly
        """
        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms', json={})
            self.forms.list(page=2, pageSize=10, search='hola', workspaceId='abc')

            history = m.request_history
            query = history[0].url.split('?')[1]
            params = dict(urllib.parse.parse_qs(query))

            self.assertEqual(params.pop('page')[0], '2')
            self.assertEqual(params.pop('page_size')[0], '10')
            self.assertEqual(params.pop('search')[0], 'hola')
            self.assertEqual(params.pop('workspace_id')[0], 'abc')

    def test_forms_get_correct_id(self):
        """
        get sends the correct UID
        """
        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms/'+self.formID, json={})
            self.forms.get(self.formID)

            history = m.request_history
            self.assertEqual(history[0].url, API_BASE_URL+'/forms/'+self.formID)

    def test_forms_get_sets_get_method(self):
        """
        get sets get method
        """
        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms/'+self.formID, json={})
            self.forms.get(self.formID)

            history = m.request_history
            self.assertEqual(history[0].method, 'GET')

    def test_forms_update_updates_a_form(self):
        """
        update updates a form
        """
        title = 'hola'
        result = self.forms.update(self.formID, data={
            'title': title
        })

        self.assertEqual(result.get('title'), title)

    def test_forms_update_as_patch_updates_a_form(self):
        """
        update as patch updates a form
        """
        result = self.forms.update(self.formID, patch=True, data=[{
            'op': 'replace',
            'path': '/title',
            'value': 'aloha'
        }])

        self.assertEqual(result, 'OK')

    def test_forms_update_sets_put_method_in_request_by_default(self):
        """
        update sets put method in request by default
        """
        with requests_mock.mock() as m:
            m.put(API_BASE_URL+'/forms/'+self.formID, json={})
            self.forms.update(self.formID, data={
                'title': 'title'
            })

            history = m.request_history

            self.assertEqual(history[0].method, 'PUT')

    def test_forms_delete_removes_the_correct_uid_form(self):
        """
        delete removes the correct uid form
        """
        get1Result = self.forms.get(self.formID)
        self.assertEqual(get1Result.get('id'), self.formID)
        self.forms.delete(self.formID)
        try:
            self.forms.get(self.formID)
        except Exception as err:
            error = str(err)
        self.assertEqual(error, 'Non existing form with uid %s' % self.formID)

    def test_forms_create_has_the_correct_path_and_method(self):
        """
        create has the correct path and method
        """
        with requests_mock.mock() as m:
            m.post(API_BASE_URL+'/forms', json={})
            self.forms.create({
                'title': 'hola'
            })

            history = m.request_history

            self.assertEqual(history[0].method, 'POST')
            self.assertEqual(history[0].url, API_BASE_URL+'/forms')

    def test_forms_create_creates_a_new_form(self):
        """
        create creates a new form
        """
        createResult = self.forms.create({
            'title': 'hola'
        })

        formID = createResult.get('id')

        getResult = self.forms.get(formID)

        self.assertIsNone(createResult.get('code', None))
        self.assertEqual(getResult.get('id'), formID)

    def test_forms_get_messages_has_the_correct_path_and_method(self):
        """
        get messages has the correct path and method
        """
        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms/'+self.formID+'/messages', json={})
            self.forms.messages.get(self.formID)

            history = m.request_history

            self.assertEqual(history[0].method, 'GET')
            self.assertEqual(history[0].url, API_BASE_URL+'/forms/'+self.formID+'/messages')

    def test_forms_update_messages_has_the_correct_path_and_method(self):
        """
        update messages has the correct path and method
        """
        with requests_mock.mock() as m:
            m.put(API_BASE_URL+'/forms/'+self.formID+'/messages')
            self.forms.messages.update(self.formID)

            history = m.request_history

            self.assertEqual(history[0].method, 'PUT')
            self.assertEqual(history[0].url, API_BASE_URL+'/forms/'+self.formID+'/messages')



### HANDLING HTML TEXT ###

import logging
import json
from requests import HTTPError, request
from bs4 import BeautifulSoup
from lxml import html
from typing import Optional, Union, List, Dict
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.remote.webelement import WebElement


class HtmlHndler:

    def html_bs_parser(self, url:str, bl_verify:bool=True,
                       method:str='GET', parser:str='html.parser') -> Union[BeautifulSoup, str]:
        '''
        DOCSTRING: HTML PARSER THROUGH BEAUTIFULSOUP
        INPUTS: HTML TEXT
        OUTPUTS: SOUP
        '''
        try:
            html_status_invest = request(method, url, verify=bl_verify).content
            return BeautifulSoup(html_status_invest, parser)
        except HTTPError as e:
            return 'HTTP Error: {}'.format(e)

    def html_lxml_parser(self, url: Optional[str]=None, page:Optional[bytes]=None,
                         method:str='GET', bl_verify:bool=True) -> html.HtmlElement:
        '''
        DOCSTRING: HTML PARSER FOR LXML PURPOSES
        INPUTS: URL, METHOD (GET AS DEFAULT) AND BOOLEAN VERIFY (TRUE AS DEFAULT)
        OUTPUTS: DOCUMENT WITH HTML CONTENT
        '''
        if page is None:
            req_resp = request(method, url, verify=bl_verify)
            req_resp.raise_for_status()
            page = req_resp.content
        return html.fromstring(page)

    def html_lxml_xpath(self, html_content, str_xpath):
        '''
        DOCSTRING: XPATH TO HANDLE LXML PARSER
        INPUTS: HTML CONTENT AND STRING XPATH
        OUTPUTS: XPATH CONTENT
        '''
        return html_content.xpath(str_xpath)

    def html_to_txt(self, html_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        soup = BeautifulSoup(html_, features='lxml')
        return soup(html_)

    def parse_html_to_string(self, html_, parsing_lib='html.parser',
                             str_body_html='',
                             join_td_character='|', td_size_ajust_character=' '):
        '''
        DOCSTRING: PARSE HTML BODY
        INPUTS: HTML
        OUTPUTS: STRING
        '''
        # setting variables
        list_ = list()
        list_tr_html = list()
        dict_ = dict()
        dict_fill_blanks_td = dict()
        # creating a parseable object
        obj_soup = BeautifulSoup(html_, parsing_lib)
        html_parsed_raw = obj_soup.get_text()
        # creating a raw parsed html body
        list_body_html = html_parsed_raw.split('\n')
        # looping through tables and periods in the raw parsed html body
        for str_ in list_body_html:
            #   append to tr, provided the value is different from empty, what is an indicative of
            #       line scape
            if str_ != '':
                list_.append(str_)
            else:
                if len(list_) > 0:
                    list_tr_html.append(list_)
                list_ = list()
            #   add tr to the list, without reseting the intermediary list, provided it is the
            #       last instance of list body html
            if (str_ == list_body_html[-1]) and (len(list_) > 0):
                list_tr_html.append(list_)
        # looping through each tr to find the maximum td length
        for i in range(len(list_tr_html)):
            #   if tr length is greater than 1 its a sign of a row from a table, otherwise its is
            #   considered a period from a phrase
            if len(list_tr_html[i]) > 1:
                dict_[i] = {j: len(list_tr_html[i][j])
                            for j in range(len(list_tr_html[i]))}
        # build dictionary with blank spaces, aiming to reach columns of same size
        for _, dict_j in dict_.items():
            for j, _ in dict_j.items():
                dict_fill_blanks_td[j] = max([dict_[i][j]
                                              for i in list(dict_.keys()) if i in dict_ and j in
                                              dict_[i]])
        # joining td's with a separator
        for i in range(len(list_tr_html)):
            #   filling blanks to construct columns of the same size
            str_body_html += join_td_character.join([list_tr_html[i][j]
                                                     + td_size_ajust_character *
                                                     (dict_fill_blanks_td[j] -
                                                      len(list_tr_html[i][j]))
                                                     for j in range(len(list_tr_html[i]))])
            #   adding line scapes
            try:
                if len(list_tr_html[i]) == len(list_tr_html[i + 1]):
                    str_body_html += '\n'
                else:
                    str_body_html += 2 * '\n'
            except IndexError:
                continue
        # returning html body parsed
        return str_body_html


class SeleniumWD:

    def __init__(self, url:str, path_webdriver:str, 
                 int_port:int, str_user_agent:str='Mozilla/5.0 (Windowns NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36', 
                 int_wait_load:int=10, 
                 int_delay:int=10,
                 bl_opn_min:bool=False, 
                 bl_headless:bool=False,
                 bl_incognito:bool=False,
                 dict_args:Optional[List[str]]=None) -> None:
        '''
        REFERENCES: 
            1. LIST OF CHROMIUM COMMAND LINE SWITCHES: https://gist.github.com/dodying/34ea4760a699b47825a766051f47d43b
            2. LIST OF USER AGENTS: https://gist.github.com/pzb/b4b6f57144aea7827ae4
        DOCSTRING: SELENIUM INTEGRATION
        INPUTS: 
            - URL:STR
            - PATH_WEBDRIVER:STR
            - PORT:INT
            - USER_AGENT:STR
            - WAIT_LOAD:INT (SECONDS)
            - DELAY:INT (SECONDS)
            - BL OPEN MINIMIZED:BOOLEAN (DEFAULT TRUE)
            - DICT ARGUMENTS:DICT (DEFAULT NONE)
        OUTPUTS: NONE
        '''
        self.url = url
        self.path_webdriver = path_webdriver
        self.int_port = int_port
        self.str_user_agent = str_user_agent
        self.int_wait_load = int_wait_load
        self.int_delay = int_delay
        self.bl_opn_min = bl_opn_min
        self.bl_headless = bl_headless
        self.bl_incognito = bl_incognito
        self.dict_default_args = dict_args if dict_args is not None else [
            '--no-sandbox',
            '--disable-gpu',
            '--disable-setuid-sandbox',
            '--disable-web-security',
            '--disable-dev-shm-usage',
            '--memory-pressure-off',
            '--ignore-certificate-errors',
            '--disable-features=site-per-process',
            f'--user-agent={str_user_agent}'
        ]
        # set headless mode for operations without graphical user interface (GUI) - if true
        if self.bl_headless == True:
            self.dict_default_args.append('--headless')
        if self.bl_incognito == True:
            self.dict_default_args.append('--incognito')
        self.browser = self.get_browser
    
    @property
    def get_browser(self) -> WebDriver:
        '''
        DOCSTRING: BROWSER INITIATION WITH THE RESPECTIVE WEB DRIVER
        INPUTS: -
        OUTPUTS: WEB DRIVER
        '''
        # setting preferences
        d = DesiredCapabilities.CHROME
        d['goog:loggingPrefs'] = {'performance': 'ALL'}
        # instantiate the browser command with passed args
        browser_options = webdriver.ChromeOptions()
        for arg in self.dict_default_args:
            browser_options.add_argument(arg)
        # instantiate the browser command
        browser = webdriver.Chrome(
            executable_path=self.path_webdriver, port=self.int_port, options=browser_options)
        # open minimized
        if \
            (self.bl_opn_min == True) \
            and (self.bl_headless == False):
            browser.minimize_window()
        # open the provided url
        browser.get(self.url)
        browser.implicitly_wait(self.int_wait_load)
        # return browser for webbrowser tests
        return browser

    def process_log(self, log:Dict[str, Union[str, dict]]) -> Optional[Dict[str, Union[str, dict]]]:
        '''
        DOCSTRING: COLLECT NETWORK ACTIVITY
        INPUTS: DRIVER, LOG
        OUTPUTS: OBJECT
        '''
        log = json.loads(log['message'])['message']
        if ('Network.response' in log['method'] and 'params' in log.keys()):
            body = self.browser.execute_cdp_cmd('Network.getResponseBody', {'requestId': log[
                'params']['requestId']})
            print(json.dumps(body, indent=4, sort_keys=True))
            return log['params']

    @property
    def get_browser_log_entries(self) -> List[Dict[str, Union[str, dict]]]:
        '''
        REFERENCES: https://stackoverflow.com/questions/20907180/getting-console-log-output-from-chrome-with-selenium-python-api-bindings
        DOCSTRING: GET LOGGING SELENIUM RESPONSE FROM WEBDRIVER
        INPUTS: DRIVER
        OUPUTS: BROWSER LOG ENTRIES (LIST)
        '''
        loglevels = {'NOTSET': 0, 'DEBUG': 10, 'INFO': 20,
                     'WARNING': 30, 'ERROR': 40, 'SEVERE': 40, 'CRITICAL': 50}
        # initialise a logger
        browserlog = logging.getLogger("chrome")
        # get browser logs
        slurped_logs = self.browser.get_log('browser')
        for entry in slurped_logs:
            # convert broswer log to python log format
            rec = browserlog.makeRecord("%s.%s" % (browserlog.name, entry['source']), loglevels.get(
                entry['level']), '.', 0, entry['message'], None, None)
            # log using original timestamp.. us -> ms
            rec.created = entry['timestamp'] / 1000
            try:
                # add browser log to python log
                browserlog.handle(rec)
            except:
                print(entry)
        # and return logs incase you want them
        return slurped_logs

    def process_browser_log_entry(self, entry: Dict[str, Union[str, dict]]) \
        -> Dict[str, Union[str, dict]]:
        '''
        REFERENCES: https://stackoverflow.com/questions/52633697/selenium-python-how-to-capture-network-traffics-response
        INPUTS: ENTRY
        OUTPUTS: STRING
        '''
        # process json from entry
        response = json.loads(entry['message'])['message']
        # return messge
        return response

    @property
    def get_network_traffic(self) -> List[Dict[str, Union[str, dict]]]:
        '''
        REFERENCES: https://stackoverflow.com/questions/52633697/selenium-python-how-to-capture-network-traffics-response
        INPUTS:
        OUTPUTS:
        '''
        # get browser log
        browser_log = self.browser.get_log('performance')
        # getting events
        list_events = [self.process_browser_log_entry(
            entry) for entry in browser_log]
        list_events = [
            event for event in list_events if 'Network.response' in event['method']]
        # returning data
        return list_events

    def find_element(self, str_element_interest:str, selector_type:str='XPATH') \
        -> WebElement:
        '''
        DOCSTRING: FINDING ELEMENT IN HTML BY SELECTOR TYPE
        INPUTS: WEB DRIVER (FROM SELENIUM, EITHER CHROME, SAFARI, FIREFOX, INTERNET EXPLORER WEB
            BROWSERS), STRING WITH THE ELEMENT OF INTEREST (IDENTIFIER TO SELECTOR),
            SELECTOR TYPE(CLAS_NAME, CSS_SELECTOR, ID, LINK_TEXT, NAME,
            PARTIAL_LINK_TEXT, TAG_NAME AND XPATH, THE FORMER AS DEFAULT)
        OUTPUTS: WEB DRIVER ELEMENT OF INTEREST
        '''
        try:
            return self.browser.find_element(getattr(By, selector_type),
                                           str_element_interest)
        except AttributeError:
            raise Exception('Attribute, difined in selector type, not available in find element, ' +
                            'please consider revisiting these argument')

    def find_elements(self, str_xpath:str) -> List[WebElement]:
        '''
        DOCSTRING: FINDING ELEMENTS IN HTML BY XPATH
        INPUTS: WEB DRIVER (FROM SELENIUM, EITHER CHROME, SAFARI, FIREFOX, INTERNET EXPLORER WEB
            BROWSERS) AND XPATH
        OUTPUTS: WEB DRIVER ELEMENT OF INTEREST
        '''
        try:
            return self.browser.find_elements_by_xpath(str_xpath)
        except AttributeError:
            raise Exception('Attribute, difined in selector type, not available in find element, ' +
                            'please consider revisiting these argument')

    def fill_input(self, web_element:WebElement, str_input:str) -> str:
        '''
        DOCSTRING: FILLING INPUT BOXES IN HTML
        INPUTS: WEB ELEMENT AND STRING TO INPUT
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        '''
        try:
            web_element.send_keys(str_input)
            return 'OK'
        except:
            raise Exception(
                'Web element error, please consider revisiting this parameter')

    def el_is_enabled(self, str_xpath:str) -> ec.ExpectedCondition:
        '''
        REFERENCES: https://github.com/clemfromspace/scrapy-selenium
        DOCSTRING: CHECK WHETHER WEB ELEMENT IS ENABLED OR NOT
        INPUTS: STR XPATH
        OUTPUTS: BOOLEAN
        '''
        return ec.element_to_be_clickable((By.XPATH, str_xpath))

    def wait_until_el_loaded(self, str_xpath:str) -> WebDriverWait:
        '''
        REFERENCES: https://stackoverflow.com/questions/26566799/wait-until-page-is-loaded-with-selenium-webdriver-for-python
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return WebDriverWait(self.browser, self.int_delay).until(self.el_is_enabled(str_xpath))


class HtmlBuilder:

    def tag(self, name, *content, cls=None, **attrs):
        '''
        REFERENCES: - FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.
        DOCSTRINGS: HTML TAG CONSTRUCTOR
        INPUTS: *ARGUMENTS, AND **ATTRIBUTES, BESIDE A CLS WORKAROUND SINCE CLASS IS A SPECIAL 
            WORD FOR PYTHON
        OUTPUTS: STRING
        '''
        # defining tag & method
        if cls is not None:
            attrs['class'] = cls
        if attrs:
            attr_str = ''.join(' {}="{}"'.format(attr, value) for attr, value
                               in sorted(attrs.items()))
        else:
            attr_str = ''
        # defining element
        if content:
            return '\n'.join('<{}{}>{}</{}>'.format(name, attr_str, c,
                                                    name) for c in content)
        else:
            return '<{}{} />'.format(name, attr_str)



FONTS_AVAILABLE = [
  'Acme',
  'Arial',
  'Arvo',
  'Avenir Next',
  'Bangers',
  'Cabin',
  'Cabin Condensed',
  'Courier',
  'Crete Round',
  'Dancing Script',
  'Exo',
  'Georgia',
  'Handlee',
  'Helvetica Neue',
  'Karla',
  'Lato',
  'Lekton',
  'Lobster',
  'Lora',
  'McLaren',
  'Montserrat',
  'Nixie One',
  'Old Standard TT',
  'Open Sans',
  'Oswald',
  'Playfair Display',
  'Quicksand',
  'Raleway',
  'Signika',
  'Sniglet',
  'Source Sans Pro',
  'Vollkorn'
]

API_BASE_URL = 'https://api.typeform.com'



### METHODS TO HANDLE STRINGS ###

import json
import uuid
import re
import ftfy
from base64 import b64encode
from basicauth import encode
from unidecode import unidecode
from unicodedata import normalize, combining
from string import ascii_uppercase, ascii_lowercase, digits
from fnmatch import fnmatch
from bs4 import BeautifulSoup


class StrHandler:

    @property
    def multi_map_reference(self):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: TRANSFORM SOME WESTERN TYPOGRAPHICAL SYMBOLS INTO ASCII, BUILDING MAPPING TABLE 
            FOR CHAR-TO-CHAR REPLACEMENT
        INPUTS: -
        OUTPUT: DICT
        '''
        single_map = str.maketrans(
            """‚ƒ„†ˆ‹‘’“”•–—˜›""", """'f"*^<''""---~>""")
        multi_map = str.maketrans({
            '€': '<euro>',
            '…': '...',
            'OE': 'OE',
            '™': '(TM)',
            'oe': 'oe',
            '‰': '<per mille>',
            '‡': '**',
        })
        return multi_map.update(single_map)

    def find_between(self, s, first, last):
        '''
        DOCSTRING: FIND STRINGS BETWEEN TWO SUBSTRINGS
        INPUTS: ORIGINAL STRING, INITAL AND FINAL DELIMITERS
        OUTPUTS: MID STRING
        '''
        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""

    def find_substr_str(self, str_, substr_):
        '''
        DOCSTRING: FINDING A SUBSTRING IN A STRING
        INPUTS: STRING AND SUBSTRING
        OUTPUTS: BOOLEAN
        '''
        return substr_ in str_

    def match_string_like(self, str_, str_like):
        '''
        DOCSTRING: MATCHING STRING WITH RESPECTIVELY STRING LIKE
        INPUTS: STRING AND STRING LIKE
        OUTPUTS: BOOLEAN
        '''
        return fnmatch(str_, str_like)

    def latin_characters(self, str_):
        '''
        DOCSTRING: CORRECTING SPECIAL CHARACTERS
        INPUTS: STRING
        OUTPUTS: CORRECTED STRING
        '''
        return str_.encode('latin1').decode('utf-8')

    def decode_special_characters_ftfy(self, str_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return ftfy.fix_str_(str_)

    def removing_accents(self, str_):
        '''
        DOCSTRING: REMOVE ACCENTS FROM LATIN ALPHABET
        INPUTS: STRING
        OUTPUTS: STRING
        '''
        return unidecode(str_)

    def byte_to_latin_characters(self, str_):
        '''
        DOCSTRING: CORRECTING SPECIAL CHARACTERS
        INPUTS: STRING
        OUTPUTS: CORRECTED STRING        
        '''
        return str_.encode('latin1').decode('ISO-8859-1')

    def remove_diacritics(self, str_):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 
        LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSITRNGS: REMOVE ALL DIACRITICS FROM A STRING, SUCH AS ACCENTS, CEDILLAS, ETC, FROM LATIN 
            AND NON-LATIN ALPHABET, LIKE GREEK.
        INPUTS: STRING
        OUTPUTS: STRING
        '''
        norm_txt = normalize('NFD', str_)
        shaved = ''.join(c for c in norm_txt if not combining(c))
        return normalize('NFC', shaved)

    def remove_end_period_marks(self, corpus, patterns='[!.?+]'):
        '''
        DOCSTRING: REMOVE END PERIOD MARKS
        INPUTS: CORPUS AND PATTERNS (DEFAULT)
        OUTPUTS: STRING
        '''
        return re.sub(patterns, '', corpus)

    def remove_only_latin_diacritics(self, str_, latin_base=False):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 
            LUCIANO RAMALHO, 978-1-491-94600-8.”        
        DOCSTRING: REMOVE ALL DISCRITIC MARKS FROM LATIN BASE CHARACTERS
        INPUTS: STRING, LATIN BASE (FALSE AS DEFAULT)
        OUTPUTS: STRING
        '''
        norm_txt = normalize('NFD', str_)
        keepers = []
        for c in norm_txt:
            if combining(c) and latin_base:
                continue  # ignore diacritic on Latin base char
            keepers.append(c)
            # if it isn't combining char, it's a new base char
            if not combining(c):
                latin_base = c in str_.ascii_letters
        shaved = ''.join(keepers)
        return normalize('NFC', shaved)

    def dewinize(self, str_):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”        
        DOCSTRING: REPLACE WIN1252 SYMBOLS WITH ASCII CHARS OR SEQUENCES
        INPUTS: STRING
        OUTPUTS: STRING
        '''
        return str_.translate(self.multi_map_reference)

    def asciize(self, str_):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”        
        DOCSTRING: APPLY NFKC NORMALIZATION TO COMPOSE CHARACTERS WITH THEIR COMPATIBILITY CODE 
            POINTS IN ASCII SYSTEM
        INPUTS: STRING
        OUTPUTS: STRING
        '''
        no_marks = self.remove_only_latin_diacritics(self.dewinize(str_))
        no_marks = no_marks.replace('ß', 'ss')
        return normalize('NFKC', no_marks)

    def remove_substr(self, str_, substr_):
        '''
        DOCSTRING: REMOVE A SUBSTRING FROM A GIVEN STRING
        INPUTS: STRING AND SUBSTRING
        OUTPUTS: STRING WITHOUT SUBSTRING
        '''
        return str_.replace(substr_, '')

    def get_string_until_substr(self, str_, substring):
        '''
        DOCSTRING: RETURN A STRING UNTIL FIND ITS SUBSTRING
        INPUTS: STRING, SUBSTRING
        OUTPUTS: STRING
        '''
        return str_.split(substring)[0]

    def get_string_after_substr(self, str_, substring):
        '''
        DOCSTRING: RETURN A STRING AFTER FIND ITS SUBSTRING
        INPUTS: STRING, SUBSTRING
        OUTPUTS: STRING
        '''
        return str_.split(substring)[1]

    def base64_encode(self, userid, password):
        '''
        DOCSTRING: ENCODING IN BASE 64 AN USER AND PASSWORD COMBINATION
        INPUTS: STRING TO ENCODE
        OUTPUTS: STRING ENCODED IN BASE64
        '''
        return encode(userid, password)

    def base64_str_encode(self, str_, code_method='ascii'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # encode to bytes
        message_bytes = str_.encode(code_method)
        # encode bytes to base64
        base64_bytes = b64encode(message_bytes)
        base64_message = base64_bytes.decode(code_method)
        # return message
        return base64_message

    @property
    def universally_unique_identifier(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # make a new uuid
        uuid_identifier = uuid.uuid4()
        # return uudi
        return {
            'uuid': uuid_identifier,
            'uuid_hex_digits_str': str(uuid_identifier),
            'uuid_32_character_hexadecimal_str': uuid_identifier.hex
        }

    def letters_to_numbers(self, letters_in_alphabet=21, first_letter_alphabet='f',
                           list_not_in_range=['i', 'l', 'o', 'p', 'r', 's', 't', 'w', 'y']):
        '''
        DOCSTRING: JSON CORRELATING LETTERS AND NUMBERS
        INPUTS: LETTERS IN ALPHABET FROM THE FIRST ONE (21 AS DEFAULT), 
            FIRST LETTER IN ALPHABET (F AS DEFAULT), LIST NOT IN RANGE (I, L, O, P, R, S, T, W, Y 
            AS DEFAULT)
        OUTPUTS: JSON WITH LETTERS IN LOWER CASE AS KEYS
        '''
        # auxiliary variables
        dict_message = dict()
        i_aux = 0

        # dictionary correlating letters and numbers
        for i in range(ord(first_letter_alphabet), ord(first_letter_alphabet) + letters_in_alphabet):
            if chr(i) not in list_not_in_range:
                dict_message[chr(i)] = i - 101 - i_aux
            else:
                i_aux += 1

        # json to export
        return json.loads(json.dumps(dict_message))

    def alphabetic_range(self, case='upper'):
        '''
        DOCSTRING: ALPHABETIC RANGE IN UPPER OR LOWER CASE
        INPUTS: CASE
        OUTPUTS: LIST
        '''
        if case == 'upper':
            return list(ascii_uppercase)
        elif case == 'lower':
            return list(ascii_lowercase)
        else:
            raise Exception('Case ought be upper or lower, although {} was given, '.format(case)
                            + 'please revisit the case variable')

    def regex_match_alphanumeric(self, str_, regex_match='^[a-zA-Z0-9_]+$'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return re.match(regex_match, str_)

    def bl_has_numbers(self, str_):
        '''
        DOCSTRING: CHECK WHETER THE STRING HAS NUMBERS
        INPUTS:
        OUTPUTS: BOOLEAN
        '''
        return bool(re.search(r'\d', str_))

    def nfc_equal(self, str1, str2):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: UNICODE EQUIVALENCE TO IDENTIFY ENCODING STARDARDS THAT REPRESENT ESSENTIALLY 
            THE SAME CHARACTER
        INPUTS: STRING 1 AND 2
        OUTPUTS: BOOLEAN
        '''
        return normalize('NFC', str1) == normalize('NFC', str2)

    def casefold_equal(self, str1, str2):
        '''
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: UNICODE EQUIVALENCE TO IDENTIFY ENCODING STARDARDS THAT REPRESENT ESSENTIALLY 
            THE SAME CASEFOLD FOR A GIVEN CHARACTER
        INPUTS: STRING 1 AND 2
        OUTPUTS: BOOLEAN
        '''
        return normalize('NFC', str1).casefold() == normalize('NFC', str2).casefold()

    def remove_non_alphanumeric_chars(self, str_, str_pattern_maintain='[\W_]',
                                      str_replace=''):
        '''
        DOCSTRING: REMOVE NON-ALPHANUMERIC CHARACTERS
        INPUTS: STRING TO BE REPLACED, STRING PATTERN ('[\W_]'  == [^a-zA-Z0-9_], AS DEFAULT), 
            AND STR REPLACE
        OUTPUTS: STRING
        '''
        return re.sub(str_pattern_maintain, str_replace, str_)

    def remove_numeric_chars(self, str_):
        '''
        REFERENCES: https://stackoverflow.com/questions/12851791/removing-numbers-from-str_
        DOCSTRING: REMOVE NUMERIC CHARACTERS
        INPUTS: STRING
        OUTPUTS: STRING
        '''
        def_remove_digits = str.maketrans('', '', digits)
        return str_.translate(def_remove_digits)

    def is_capitalized(self, str_, bl_simple_validation=True):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        # removing diacritcs
        str_ = self.remove_diacritics(str_)
        # removing non-alfanumeric characters
        str_ = self.remove_non_alphanumeric_chars(str_)
        #   returning wheter is capitalized or not
        try:
            if bl_simple_validation == True:
                if (str_[0].isupper() == True) and (str_[1].islower() == True):
                    return True
                else:
                    return False
            else:
                if (str_[0].isupper() == True) and (all([l.islower() for l in str_[1:]])):
                    return True
                else:
                    return False
        except:
            return False

    def split_re(self, str_, re_split=r'[;,\s]\s*'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return re.split(re_split, str_)

    def replace_case_insensitive(self, str_, str_replaced, str_replace):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return re.sub(str_replaced, str_replace, str_, flags=re.IGNORECASE)

    def matchcase(str_):
        '''
        REFERENCES: PYTHON COOKBOOK - DAVID BEASZLEY, BRIAN K. JONES
        DOCSTRING: MATCHASE SENSE
        INPUTS: WORD
        OUTPUTS: STRING
        '''
        def replace(m):
            str_ = m.group()
            if str_.isupper():
                return str_.upper()
            elif str_.islower():
                return str_.lower()
            elif str_[0].isupper():
                return str_.capitalize()
            else:
                return str_
        return replace

    def replace_respecting_case(self, str_, str_replaced, str_replace):
        '''
        REFERENCES: PYTHON COOKBOOK - DAVID BEASZLEY, BRIAN K. JONES
        DOCSTRING: MATCHASE SENSE
        INPUTS: WORD
        OUTPUTS: STRING
        '''
        return re.sub(str_replaced, self.matchcase(str_replace), str_, flags=re.IGNORECASE)

    def replace_all(self, str_, dict_replacers):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        for i, j in dict_replacers.items():
            str_ = str_.replace(i, j)
        return str_

    def html_to_txt(self, html_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        soup = BeautifulSoup(html_, features='lxml')
        return soup(html_)

    def extract_urls(self, str_):
        '''
        DOCSTRING: LIST OF URLS IN A GIVEN STRING
        INPUTS: STRING
        OUTPUTS: LIST
        '''
        # define a regular expression pattern to match URLs
        url_pattern = re.compile(r'https?://\S+|www\.\S+')
        # find all matches in the given str_
        list_urls = re.findall(url_pattern, str_)
        # return urls list
        return list_urls

    def is_word(self, _value):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        try:
            float(_value)
            return False
        except ValueError:
            return True
        
    def camel_to_snake(self, str_cammel):
        '''
        Cammel case to snake case
        Arguments:
            str_ (str): string to be converted
        Outputs: str
        '''
        str_snake = re.sub(r'(?<!^)(?=[A-Z])', '_', str_cammel).upper()
        return str_snake